# Reticulum License
#
# Copyright (c) 2016-2026 Mark Qvist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# - The Software shall not be used in any kind of system which includes amongst
#   its functions the ability to purposefully do harm to human beings.
#
# - The Software shall not be used, directly or indirectly, in the creation of
#   an artificial intelligence, machine learning or language model training
#   dataset, including but not limited to any use that contributes to the
#   training or development of such a model or algorithm.
#
# - The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re
import RNS

# Validate ref names according to https://git-scm.com/docs/git-check-ref-format
# This may be a bit overkill, since git validates names as well, but why not.
def san_ref(ref):
    if ref.startswith("-"):                return None
    if ref.startswith("/"):                return None
    if ref.endswith("/"):                  return None
    if ref.endswith("."):                  return None

    if " "     in ref:                     return None
    if not "/" in ref:                     return None
    if ".."    in ref:                     return None
    if "/."    in ref:                     return None
    if "//"    in ref:                     return None
    if "\\"    in ref:                     return None

    for comp in ref.split("/"):
        if comp.endswith(".lock"):         return None

    if not all(ord(c) >= 40 for c in ref): return None # Any control character
    if "\x7f" in ref:                      return None # ASCII DEL (177)
    if "~"    in ref:                      return None
    if "^"    in ref:                      return None
    if ":"    in ref:                      return None
    if "?"    in ref:                      return None
    if "*"    in ref:                      return None
    if "["    in ref:                      return None
    if "@{"   in ref:                      return None
    if "@"    == ref:                      return None

    return ref

def san_refs(refs):
    if not type(refs) == list: return None
    for ref in refs:
        if not san_ref(ref): return None

    return refs

# Git SHA format validation
def san_sha(sha):
    if len(sha) < 40: return None
    try: bytes.fromhex(sha)
    except: return None
    return sha

class MarkdownToMicron:    
    BOLD = "`!"
    BOLD_END = "`!"
    ITALIC = "`*"
    ITALIC_END = "`*"
    UNDERLINE = "`_"
    UNDERLINE_END = "`_"
    
    CODE_BG = "`BT282828"
    CODE_BG_INLINE = "`BT383838"
    CODE_FG = "`Fddd"
    CODE_RESET = "`f`b"
    
    LITERAL_START = "`="
    LITERAL_END = "`="
    
    BULLET = "•"
    
    # Regex patterns for markdown elements
    HEADER_RE = re.compile(r'^(#{1,6})\s+(.+)$')
    CODE_FENCE_RE = re.compile(r'^(\s*)```(.*)$')
    HORIZONTAL_RULE_RE = re.compile(r'^(\s*)(---+|===+|\*\*\*+|___+)\s*$')
    UNORDERED_LIST_RE = re.compile(r'^(\s*)([-*+])\s+(.+)$')
    
    # Table patterns
    TABLE_ROW_RE = re.compile(r'^\s*\|?(.+?)\|?\s*$')
    TABLE_SEP_RE = re.compile(r'^\s*\|?(?:\s*:?-+:?\s*\|)+\s*$')
    
    # Quote pattern
    QUOTE_RE = re.compile(r'^>\s?(.*)$')
    
    # Inline patterns (processed in order of specificity)
    LINK_RE = re.compile(r'\[([^\]]+)\]\(([^)]+)\)')
    INLINE_CODE_RE = re.compile(r'`([^`]+)`')
    BOLD_RE = re.compile(r'\*\*(.+?)\*\*|__(.+?)__')
    ITALIC_RE = re.compile(r'\*(.+?)\*|_(.+?)_')

    TABLE_H = "─"
    TABLE_V = "│"
    TABLE_TL = "┌"
    TABLE_TR = "┐"
    TABLE_BL = "└"
    TABLE_BR = "┘"
    TABLE_ML = "├"
    TABLE_MR = "┤"
    TABLE_TM = "┬"
    TABLE_BM = "┴"
    TABLE_MM = "┼"
    
    TABLE_MIN_COL_WIDTH = 3

    def __init__(self, max_width=100, syntax_highlighter=None, url_scope=None):
        self.max_width = max_width
        self.local_url_scope = url_scope or ":/page/"
        self.__local_url_scope = self.local_url_scope
        self.syntax_highlighter = syntax_highlighter
        self.wcwidth = None

        self.bold_links = True
        self.underline_links = True
        self.link_color = None
        
        try:
            import wcwidth
            self.wcwidth = wcwidth

        except: RNS.log(f"The wcwidth module is unavailable, display width calculations for some glyphs will be incorrect", RNS.LOG_WARNING)

    def set_url_scope(self, url_scope): self.local_url_scope = url_scope
    def restore_url_scope(self, url_scope): self.local_url_scope = self.__local_url_scope

    def display_width(self, text):
        if not self.wcwidth: return len(text)
        else:
            # wcswidth returns -1 for non-printable strings,
            # fallback to len in this case
            w = self.wcwidth.wcswidth(text)
            return w if w is not None and w >= 0 else len(text)
    
    def format_block(self, text, url_scope=None):
        # text = text.replace("\\", "\\\\") # Now handled in format_line instead
        lines = text.split('\n')
        result_lines = []
        in_code_block = False
        code_block_lang = None
        code_buffer = []
        in_table = False
        table_buffer = []
        in_quote = False
        quote_buffer = []
        
        def flush_quote_buffer():
            nonlocal result_lines, quote_buffer, in_quote
            if not quote_buffer:
                in_quote = False
                return
            
            para = " ".join(quote_buffer)
            formatted = self._format_inline(para)
            
            effective_width = self.max_width - 3
            if effective_width < 1: effective_width = 1
            wrapped_lines = self._wrap_text(formatted, effective_width)
            for wrapped_line in wrapped_lines: result_lines.append(f" │ {wrapped_line}")
            
            quote_buffer = []
            in_quote = False
        
        def flush_table_buffer():
            nonlocal result_lines, table_buffer, in_table
            if not table_buffer:
                in_table = False
                return
            
            if len(table_buffer) >= 2 and self._is_table_separator(table_buffer[1]):
                formatted_lines = self.format_table(table_buffer)
                result_lines.extend(formatted_lines)
            
            else:
                for line in table_buffer: result_lines.append(self.format_line(line))
            
            table_buffer = []
            in_table = False
        
        def flush_code_block():
            nonlocal result_lines, code_buffer, code_block_lang
            if not code_buffer:
                return
            
            code_content = '\n'.join(code_buffer)
            
            if self.syntax_highlighter and code_block_lang:
                if code_block_lang.lower() == "rawmu": result_lines.append(code_content)
                else:
                    try:
                        highlighted = self.syntax_highlighter.highlight(code_content, language=code_block_lang)
                        result_lines.append(f"{self.CODE_BG}{self.CODE_FG}")
                        result_lines.append(highlighted)
                        result_lines.append(self.CODE_RESET)

                    except Exception:
                        # Fallback to plain literal block on any error
                        result_lines.append(f"{self.CODE_BG}{self.CODE_FG}")
                        result_lines.append(self.LITERAL_START)
                        result_lines.append(self._escape_literals(code_content))
                        result_lines.append(self.LITERAL_END)
                        result_lines.append(self.CODE_RESET)
            else:
                result_lines.append(f"{self.CODE_BG}{self.CODE_FG}")
                result_lines.append(self.LITERAL_START)
                result_lines.append(self._escape_literals(code_content))
                result_lines.append(self.LITERAL_END)
                result_lines.append(self.CODE_RESET)
            
            code_buffer = []
        
        for line in lines:
            is_fence, lang_hint = self._detect_code_fence(line)

            if is_fence:
                # Flush any pending structures before code fence
                flush_quote_buffer()
                flush_table_buffer()
                
                if not in_code_block:
                    # Opening fence, start buffering
                    in_code_block = True
                    code_block_lang = lang_hint.strip() if lang_hint else None
                    code_buffer = []
                
                else:
                    # Closing fence, flush highlighted code
                    flush_code_block()
                    in_code_block = False
                    code_block_lang = None
            
            else:
                # Buffer code lines for later highlighting
                if in_code_block: code_buffer.append(line)
                else:
                    quote_match = self.QUOTE_RE.match(line)
                    if quote_match:
                        if not in_quote:
                            flush_table_buffer()
                            in_quote = True
                            quote_buffer = []
                        
                        quote_buffer.append(quote_match.group(1))
                    
                    else:
                        if in_quote:
                            flush_quote_buffer()
                            if line.strip() != "":
                                if self._is_table_row(line):
                                    in_table = True
                                    table_buffer = [line]
                                
                                else:
                                    formatted = self.format_line(line)
                                    result_lines.append(formatted)
                            
                            # Pass through blank line as separator
                            else: result_lines.append("")
                        
                        else:
                            if self._is_table_row(line):
                                if not in_table:
                                    in_table = True
                                    table_buffer = [line]
                                
                                else: table_buffer.append(line)
                            
                            else:
                                # Line breaks table, flush buffer
                                if in_table: flush_table_buffer()
                                formatted = self.format_line(line)
                                result_lines.append(formatted)
        
        # Handle unclosed structures
        if in_quote: flush_quote_buffer()
        if in_table: flush_table_buffer()
        if in_code_block: flush_code_block()
        
        return '\n'.join(result_lines)
    
    def format_line(self, line, mode="normal"):
        if mode == "codeblock": return self._escape_literals(line)
        line = line.replace("\\", "\\\\")

        if line.startswith("-") and not line.startswith("---") and not line.startswith("- "): line = f"\\{line}"
        if line.startswith("<"): line = f"\\{line}"
        # if line.startswith(">"): line = f"\\{line}" # Now handled by blockquotes
        
        if self.HORIZONTAL_RULE_RE.match(line): return self._format_horizontal_rule()
        
        header_match = self.HEADER_RE.match(line)
        if header_match: return self._format_header(header_match)
        
        list_match = self.UNORDERED_LIST_RE.match(line)
        if list_match: return self._format_list_item(list_match)
        
        line = self._format_inline(line)
        
        return line
    
    def _format_inline(self, text):
        code_blocks = []
        def extract_code(match):
            code_blocks.append(match.group(1))
            return f"\x00CODE{len(code_blocks)-1}\x00"

        links = []
        def extract_link(match):
            links.append((match.group(1), match.group(2)))
            return f"\x00LINK{len(links)-1}\x00"
        
        text = self.LINK_RE.sub(extract_link, text)
        text = self.INLINE_CODE_RE.sub(extract_code, text)
        text = self.BOLD_RE.sub(self._bold_sub, text)
        text = self.ITALIC_RE.sub(self._italic_sub, text)
        
        def restore_link(match):
            idx = int(match.group(1))
            text, url = links[idx]
            
            anchor_components = url.split("#")
            url = anchor_components[0]
            anchor = anchor_components[1] if len(anchor_components) > 1 else ""

            if not ":/" in url:
                url = f"{self.local_url_scope}{url}"
                if anchor: url = f"{url}|anchor={anchor}"

            undl = "`_" if self.underline_links else ""
            bold = "`!" if self.bold_links else ""
            text = text.replace('`', '')
            link = f"{undl}{bold}`[{text}`{url}]{bold}{undl}"

            if self.link_color and len(self.link_color) == 3: link = f"`F{self.link_color}{link}`f"
            if self.link_color and len(self.link_color) == 6: link = f"`FT{self.link_color}{link}`f"

            return link
        
        text = re.sub(r'\x00LINK(\d+)\x00', restore_link, text)
        
        def restore_code(match):
            idx = int(match.group(1))
            content = code_blocks[idx]
            content = content.replace('`', '\\`')
            return f"{self.CODE_BG_INLINE}{self.CODE_FG}{content}{self.CODE_RESET}"
        
        text = re.sub(r'\x00CODE(\d+)\x00', restore_code, text)
        return text
    
    def _highlight_inline_code(self, content):
        if not self.syntax_highlighter: return None
        return self.syntax_highlighter.highlight(content, language=None)
    
    def _bold_sub(self, match):
        content = match.group(1) or match.group(2)
        return f"{self.BOLD}{content}{self.BOLD_END}"
    
    def _italic_sub(self, match):
        content = match.group(1) or match.group(2)
        return f"{self.ITALIC}{content}{self.ITALIC_END}"
    
    def _format_header(self, match):
        hashes = match.group(1)
        content = match.group(2)
        level = len(hashes)
        prefix = ">" * min(level, 6)
        return f"{prefix}{self._format_inline(content)}"
    
    def _format_list_item(self, match):
        indent = match.group(1)
        content = match.group(3)
        content = self._format_inline(content)
        return f"{indent} {self.BULLET} {content}"
    
    def _format_horizontal_rule(self):
        return "-"
    
    def _detect_code_fence(self, line):
        match = self.CODE_FENCE_RE.match(line)
        if match:
            # match.group(2) contains everything after the backticks (language hint)
            return True, match.group(2)
        return False, ""
    
    def _is_table_row(self, line):
        if '|' not in line: return False
        match = self.TABLE_ROW_RE.match(line)
        if match is None: return False
        content = match.group(1)
        return '|' in content or line.strip().startswith('|')
    
    def _is_table_separator(self, line):
        if '|' not in line: return False
        match = self.TABLE_SEP_RE.match(line)
        return match is not None
    
    def _escape_literals(self, text):
        return text.replace('`', '\\`')

    def format_table(self, rows, align="c"):
        if len(rows) < 2: return rows
        
        # Parse header and separator
        header_cells = self._parse_table_row(rows[0])
        alignments = self._parse_table_alignments(rows[1])
        
        # Ensure alignment count matches header cells
        while len(alignments) < len(header_cells): alignments.append('left')
        alignments = alignments[:len(header_cells)]
        
        # Parse data rows
        data_rows = []
        for i in range(2, len(rows)):
            cells = self._parse_table_row(rows[i])
            while len(cells) < len(header_cells): cells.append("")
            cells = cells[:len(header_cells)]
            data_rows.append(cells)
        
        # Calculate column widths based on content
        num_cols = len(header_cells)
        col_widths = [0] * num_cols
        
        all_rows = [header_cells] + data_rows
        for row in all_rows:
            for i, cell in enumerate(row):
                formatted = self._format_inline(cell)
                width = self._visible_width(formatted)
                col_widths[i] = max(col_widths[i], width)
        
        # Apply minimum width and calculate total
        col_widths = [max(w, self.TABLE_MIN_COL_WIDTH) for w in col_widths]
        
        # Check max_width constraint
        # Total = sum of columns + 3 chars per column (space + 2 borders) + 1 for final border
        total_width = sum(col_widths) + (num_cols * 3) + 1
        
        if total_width > self.max_width:
            # Reduce widest columns proportionally
            excess = total_width - self.max_width
            indexed_widths = [(i, w) for i, w in enumerate(col_widths)]
            indexed_widths.sort(key=lambda x: -x[1])
            
            for i, w in indexed_widths:
                if excess <= 0: break
                reduction = min(excess, w - self.TABLE_MIN_COL_WIDTH)
                col_widths[i] -= reduction
                excess -= reduction
        
        # Build formatted table
        result = []
        
        # Alignment start
        if align: result.append(f"`{align}")
        
        # Top border
        border = self.TABLE_TL
        for i, w in enumerate(col_widths):
            border += self.TABLE_H * (w + 2)
            if i < len(col_widths) - 1: border += self.TABLE_TM
            else:                       border += self.TABLE_TR
        
        result.append(self._escape_literals(border))
        
        # Header row
        header_line = self.TABLE_V
        for i, cell in enumerate(header_cells):
            formatted = self._format_inline(cell)
            padded = self._pad_cell(formatted, col_widths[i], 'left')
            header_line += f" {padded} {self.TABLE_V}"
        result.append(self._escape_literals(header_line))
        
        # Separator row
        sep_line = self.TABLE_ML
        for i, w in enumerate(col_widths):
            cell_width = w + 2
            sep_line += self.TABLE_H * cell_width
            
            if i < len(col_widths) - 1: sep_line += self.TABLE_MM
            else:                       sep_line += self.TABLE_MR

        result.append(self._escape_literals(sep_line))
        
        # Data rows
        for row in data_rows:
            row_line = self.TABLE_V
            for i, cell in enumerate(row):
                formatted = self._format_inline(cell)
                padded = self._pad_cell(formatted, col_widths[i], alignments[i])
                row_line += f" {padded} {self.TABLE_V}"
            
            result.append(row_line)
        
        # Bottom border
        border = self.TABLE_BL
        for i, w in enumerate(col_widths):
            border += self.TABLE_H * (w + 2)
            if i < len(col_widths) - 1: border += self.TABLE_BM
            else:                       border += self.TABLE_BR
        
        result.append(self._escape_literals(border))
        
        # End alignment
        if align: result.append("`a")
        
        return result

    def format_table_raw(self, rows, align="c"):
        if len(rows) < 2: return rows
        
        # Parse header and separator
        header_cells = self._parse_table_row(rows[0])
        alignments = self._parse_table_alignments(rows[1])
        
        # Ensure alignment count matches header cells
        while len(alignments) < len(header_cells): alignments.append('left')
        alignments = alignments[:len(header_cells)]
        
        # Parse data rows
        data_rows = []
        for i in range(2, len(rows)):
            cells = self._parse_table_row(rows[i])
            while len(cells) < len(header_cells): cells.append("")
            cells = cells[:len(header_cells)]
            data_rows.append(cells)
        
        # Calculate column widths based on raw content
        num_cols = len(header_cells)
        col_widths = [0] * num_cols
        
        all_rows = [header_cells] + data_rows
        for row in all_rows:
            for i, cell in enumerate(row):
                width = self._visible_width(cell)
                col_widths[i] = max(col_widths[i], width)
        
        # Apply minimum width and calculate total
        col_widths = [max(w, self.TABLE_MIN_COL_WIDTH) for w in col_widths]
        
        # Check max_width constraint
        total_width = sum(col_widths) + (num_cols * 3) + 1
        
        if total_width > self.max_width:
            # Reduce widest columns proportionally
            excess = total_width - self.max_width
            indexed_widths = [(i, w) for i, w in enumerate(col_widths)]
            indexed_widths.sort(key=lambda x: -x[1])
            
            for i, w in indexed_widths:
                if excess <= 0: break
                reduction = min(excess, w - self.TABLE_MIN_COL_WIDTH)
                col_widths[i] -= reduction
                excess -= reduction
        
        # Build formatted table
        result = []
        
        # Alignment start
        if align: result.append(f"`{align}")
        
        # Top border
        border = self.TABLE_TL
        for i, w in enumerate(col_widths):
            border += self.TABLE_H * (w + 2)
            if i < len(col_widths) - 1: border += self.TABLE_TM
            else:                       border += self.TABLE_TR
        
        result.append(self._escape_literals(border))
        
        # Header row
        header_line = self.TABLE_V
        for i, cell in enumerate(header_cells):
            padded = self._pad_cell(cell, col_widths[i], 'left')
            header_line += f" {padded} {self.TABLE_V}"
        result.append(header_line)
        
        # Separator row - clean horizontal lines without alignment markers
        sep_line = self.TABLE_ML
        for i, w in enumerate(col_widths):
            cell_width = w + 2
            sep_line += self.TABLE_H * cell_width
            
            if i < len(col_widths) - 1: sep_line += self.TABLE_MM
            else:                       sep_line += self.TABLE_MR
        
        result.append(self._escape_literals(sep_line))
        
        # Data rows (with alignment)
        for row in data_rows:
            row_line = self.TABLE_V
            for i, cell in enumerate(row):
                padded = self._pad_cell(cell, col_widths[i], alignments[i])
                row_line += f" {padded} {self.TABLE_V}"
            
            result.append(row_line)
        
        # Bottom border
        border = self.TABLE_BL
        for i, w in enumerate(col_widths):
            border += self.TABLE_H * (w + 2)
            if i < len(col_widths) - 1: border += self.TABLE_BM
            else:                       border += self.TABLE_BR
        
        result.append(self._escape_literals(border))
        
        # End alignment
        if align: result.append("`a")
        
        return result
    
    def _parse_table_row(self, line):
        line = line.strip()
        if line.startswith('|'): line = line[1:]
        if line.endswith('|'):   line = line[:-1]
        
        cells = []
        current = ""
        escaped = False
        for char in line:
            if escaped:
                current += char
                escaped = False
            elif char == '\\':
                escaped = True
            elif char == '|':
                cells.append(current.strip())
                current = ""
            else:
                current += char
        
        cells.append(current.strip())
        return cells
    
    def _parse_table_alignments(self, line):
        cells = self._parse_table_row(line)
        alignments = []
        for cell in cells:
            cell = cell.strip()
            if cell.startswith(':') and cell.endswith(':'): alignments.append('center')
            elif cell.endswith(':'):                        alignments.append('right')
            else:                                           alignments.append('left')
        
        return alignments
    
    def _visible_width(self, text):
        text = re.sub(r'`[FB][0-9a-fA-F]{3}', '', text)
        text = re.sub(r'`[FB]T[0-9a-fA-F]{6}', '', text)
        text = re.sub(r'`[!*_=]', '', text)
        text = re.sub(r'`f`b', '', text)
        text = re.sub(r'`f', '', text)
        text = re.sub(r'`b', '', text)
        return self.display_width(text)
    
    def _pad_cell(self, text, width, align):
        text = self._truncate_cell(text, width)
        text_width = self._visible_width(text)
        padding = width - text_width
        
        if align == 'right':
            return " " * padding + text
        elif align == 'center':
            left = padding // 2
            right = padding - left
            return " " * left + text + " " * right
        else:
            return text + " " * padding

    def _truncate_cell(self, text, width):
        if self._visible_width(text) <= width: return text

        truncation_point = len(text)
        while truncation_point > 0 and self._visible_width(text[0:truncation_point]) >= width:
            truncation_point -= 1

        truncated = text[:truncation_point]

        # Yes, this is convoluted, but if someone else has
        # a better idea on how to handle unclosed micron
        # tags in the truncated cells, I'm all ears.
        active_tags = set()
        fg_active = False
        bg_active = False

        i = 0
        while i < len(truncated):
            if truncated[i] == '`':
                if i + 1 < len(truncated):
                    tag_char = truncated[i + 1]

                    if tag_char in '!*_=':
                        if tag_char in active_tags: active_tags.remove(tag_char)
                        else:                       active_tags.add(tag_char)
                        i += 2
                        continue

                    elif tag_char == 'f':
                        fg_active = False
                        i += 2
                        continue

                    elif tag_char == 'b':
                        bg_active = False
                        i += 2
                        continue

                    elif tag_char == 'F':
                        fg_active = True
                        if i + 2 < len(truncated) and truncated[i + 2] == 'T': i += 8
                        else:                                                  i += 5
                        continue

                    elif tag_char == 'B':
                        bg_active = True
                        if i + 2 < len(truncated) and truncated[i + 2] == 'T': i += 8
                        else:                                                  i += 5
                        continue
            i += 1

        closers = []
        if fg_active: closers.append('`f')
        if bg_active: closers.append('`b')
        for fmt in active_tags: closers.append(f'`{fmt}')

        return truncated + ''.join(closers) + "…"
    
    def _wrap_text(self, text, width):
        if not text: return [""]
        
        words = text.split(' ')
        lines = []
        current_line = ""
        current_width = 0
        
        for word in words:
            if not word: continue
            
            word_width = self._visible_width(word)
            
            # Check if word alone exceeds width to force break it
            if word_width > width:
                if current_line:
                    lines.append(current_line)
                    current_line = ""
                    current_width = 0
                
                # Force break the long word character by character
                remaining = word
                while remaining:
                    # Binary search for how many characters fit
                    low, high = 1, len(remaining)
                    fit_chars = 0
                    
                    while low <= high:
                        mid = (low + high) // 2
                        test_substr = remaining[:mid]
                        test_width = self._visible_width(test_substr)
                        
                        if test_width <= width:
                            fit_chars = mid
                            low = mid + 1
                        else:
                            high = mid - 1
                    
                    if fit_chars == 0: fit_chars = 1 # Need to force progress
                    
                    lines.append(remaining[:fit_chars])
                    remaining = remaining[fit_chars:]
                
                continue
            
            # Check if word fits on current line
            space_width = 1 if current_line else 0
            if current_width + space_width + word_width <= width:
                if current_line:
                    current_line += " " + word
                    current_width += space_width + word_width
                else:
                    current_line = word
                    current_width = word_width
            else:
                # Flush current line and start new one
                lines.append(current_line)
                current_line = word
                current_width = word_width
        
        # Don't forget the last line
        if current_line: lines.append(current_line)
        
        return lines if lines else [""]


def convert_markdown_to_micron(text):
    converter = MarkdownToMicron()
    return converter.format_block(text)



# Reticulum License
#
# Copyright (c) 2016-2025 Mark Qvist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# - The Software shall not be used in any kind of system which includes amongst
#   its functions the ability to purposefully do harm to human beings.
#
# - The Software shall not be used, directly or indirectly, in the creation of
#   an artificial intelligence, machine learning or language model training
#   dataset, including but not limited to any use that contributes to the
#   training or development of such a model or algorithm.
#
# - The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from RNS.Interfaces.Interface import Interface
from time import sleep
import sys
import threading
import time
import RNS

class KISS():
    FEND              = 0xC0
    FESC              = 0xDB
    TFEND             = 0xDC
    TFESC             = 0xDD
    CMD_UNKNOWN       = 0xFE
    CMD_DATA          = 0x00
    CMD_TXDELAY       = 0x01
    CMD_P             = 0x02
    CMD_SLOTTIME      = 0x03
    CMD_TXTAIL        = 0x04
    CMD_FULLDUPLEX    = 0x05
    CMD_SETHARDWARE   = 0x06
    CMD_READY         = 0x0F
    CMD_RETURN        = 0xFF

    @staticmethod
    def escape(data):
        data = data.replace(bytes([0xdb]), bytes([0xdb, 0xdd]))
        data = data.replace(bytes([0xc0]), bytes([0xdb, 0xdc]))
        return data

class AX25():
    PID_NOLAYER3    = 0xF0
    CTRL_UI         = 0x03
    CRC_CORRECT     = bytes([0xF0])+bytes([0xB8])
    HEADER_SIZE     = 16


class AX25KISSInterface(Interface):
    MAX_CHUNK = 32768
    BITRATE_GUESS = 1200
    DEFAULT_IFAC_SIZE = 8

    owner    = None
    port     = None
    speed    = None
    databits = None
    parity   = None
    stopbits = None
    serial   = None

    def __init__(self, owner, configuration):
        import importlib.util
        if importlib.util.find_spec('serial') != None:
            import serial
        else:
            RNS.log("Using the AX.25 KISS interface requires a serial communication module to be installed.", RNS.LOG_CRITICAL)
            RNS.log("You can install one with the command: python3 -m pip install pyserial", RNS.LOG_CRITICAL)
            RNS.panic()

        super().__init__()

        c = Interface.get_config_obj(configuration)
        name = c["name"]
        preamble = int(c["preamble"]) if "preamble" in c else None
        txtail = int(c["txtail"]) if "txtail" in c else None
        persistence = int(c["persistence"]) if "persistence" in c else None
        slottime = int(c["slottime"]) if "slottime" in c else None
        flow_control = c.as_bool("flow_control") if "flow_control" in c else False
        port = c["port"] if "port" in c else None
        speed = int(c["speed"]) if "speed" in c else 9600
        databits = int(c["databits"]) if "databits" in c else 8
        parity = c["parity"] if "parity" in c else "N"
        stopbits = int(c["stopbits"]) if "stopbits" in c else 1

        callsign = c["callsign"] if "callsign" in c else ""
        ssid = int(c["ssid"]) if "ssid" in c else -1

        if port == None:
            raise ValueError("No port specified for serial interface")

        self.HW_MTU = 564
        
        self.pyserial = serial
        self.serial   = None
        self.owner    = owner
        self.name     = name
        self.src_call = callsign.upper().encode("ascii")
        self.src_ssid = ssid
        self.dst_call = "APZRNS".encode("ascii")
        self.dst_ssid = 0
        self.port     = port
        self.speed    = speed
        self.databits = databits
        self.parity   = serial.PARITY_NONE
        self.stopbits = stopbits
        self.timeout  = 100
        self.online   = False
        self.bitrate  = AX25KISSInterface.BITRATE_GUESS

        self.packet_queue    = []
        self.flow_control    = flow_control
        self.interface_ready = False
        self.flow_control_timeout = 5
        self.flow_control_locked  = time.time()

        if (len(self.src_call) < 3 or len(self.src_call) > 6):
            raise ValueError("Invalid callsign for "+str(self))

        if (self.src_ssid < 0 or self.src_ssid > 15):
            raise ValueError("Invalid SSID for "+str(self))

        self.preamble    = preamble if preamble != None else 350;
        self.txtail      = txtail if txtail != None else 20;
        self.persistence = persistence if persistence != None else 64;
        self.slottime    = slottime if slottime != None else 20;

        if parity.lower() == "e" or parity.lower() == "even":
            self.parity = serial.PARITY_EVEN

        if parity.lower() == "o" or parity.lower() == "odd":
            self.parity = serial.PARITY_ODD

        try:
            self.open_port()
        except Exception as e:
            RNS.log("Could not open serial port for interface "+str(self), RNS.LOG_ERROR)
            raise e

        if self.serial.is_open:
            self.configure_device()
        else:
            raise IOError("Could not open serial port")

    def open_port(self):
        RNS.log("Opening serial port "+self.port+"...", RNS.LOG_VERBOSE)
        self.serial = self.pyserial.Serial(
            port = self.port,
            baudrate = self.speed,
            bytesize = self.databits,
            parity = self.parity,
            stopbits = self.stopbits,
            xonxoff = False,
            rtscts = False,
            timeout = 0,
            inter_byte_timeout = None,
            write_timeout = None,
            dsrdtr = False,
        )

    def configure_device(self):
        # Allow time for interface to initialise before config
        sleep(2.0)
        thread = threading.Thread(target=self.readLoop)
        thread.daemon = True
        thread.start()
        self.online = True
        RNS.log("Serial port "+self.port+" is now open")
        RNS.log("Configuring AX.25 KISS interface parameters...")
        self.setPreamble(self.preamble)
        self.setTxTail(self.txtail)
        self.setPersistence(self.persistence)
        self.setSlotTime(self.slottime)
        self.setFlowControl(self.flow_control)
        self.interface_ready = True
        RNS.log("AX.25 KISS interface configured")

    def setPreamble(self, preamble):
        preamble_ms = preamble
        preamble = int(preamble_ms / 10)
        if preamble < 0:
            preamble = 0
        if preamble > 255:
            preamble = 255

        kiss_command = bytes([KISS.FEND])+bytes([KISS.CMD_TXDELAY])+bytes([preamble])+bytes([KISS.FEND])
        written = self.serial.write(kiss_command)
        if written != len(kiss_command):
            raise IOError("Could not configure AX.25 KISS interface preamble to "+str(preamble_ms)+" (command value "+str(preamble)+")")

    def setTxTail(self, txtail):
        txtail_ms = txtail
        txtail = int(txtail_ms / 10)
        if txtail < 0:
            txtail = 0
        if txtail > 255:
            txtail = 255

        kiss_command = bytes([KISS.FEND])+bytes([KISS.CMD_TXTAIL])+bytes([txtail])+bytes([KISS.FEND])
        written = self.serial.write(kiss_command)
        if written != len(kiss_command):
            raise IOError("Could not configure AX.25 KISS interface TX tail to "+str(txtail_ms)+" (command value "+str(txtail)+")")

    def setPersistence(self, persistence):
        if persistence < 0:
            persistence = 0
        if persistence > 255:
            persistence = 255

        kiss_command = bytes([KISS.FEND])+bytes([KISS.CMD_P])+bytes([persistence])+bytes([KISS.FEND])
        written = self.serial.write(kiss_command)
        if written != len(kiss_command):
            raise IOError("Could not configure AX.25 KISS interface persistence to "+str(persistence))

    def setSlotTime(self, slottime):
        slottime_ms = slottime
        slottime = int(slottime_ms / 10)
        if slottime < 0:
            slottime = 0
        if slottime > 255:
            slottime = 255

        kiss_command = bytes([KISS.FEND])+bytes([KISS.CMD_SLOTTIME])+bytes([slottime])+bytes([KISS.FEND])
        written = self.serial.write(kiss_command)
        if written != len(kiss_command):
            raise IOError("Could not configure AX.25 KISS interface slot time to "+str(slottime_ms)+" (command value "+str(slottime)+")")

    def setFlowControl(self, flow_control):
        kiss_command = bytes([KISS.FEND])+bytes([KISS.CMD_READY])+bytes([0x01])+bytes([KISS.FEND])
        written = self.serial.write(kiss_command)
        if written != len(kiss_command):
            if (flow_control):
                raise IOError("Could not enable AX.25 KISS interface flow control")
            else:
                raise IOError("Could not enable AX.25 KISS interface flow control")


    def process_incoming(self, data):
        if (len(data) > AX25.HEADER_SIZE):
            self.rxb += len(data)
            self.owner.inbound(data[AX25.HEADER_SIZE:], self)


    def process_outgoing(self,data):
        datalen = len(data)
        if self.online:
            if self.interface_ready:
                if self.flow_control:
                    self.interface_ready = False
                    self.flow_control_locked = time.time()

                encoded_dst_ssid = bytes([0x60 | (self.dst_ssid << 1)])
                encoded_src_ssid = bytes([0x60 | (self.src_ssid << 1) | 0x01])

                addr = b""

                for i in range(0,6):
                    if (i < len(self.dst_call)):
                        addr += bytes([self.dst_call[i]<<1])
                    else:
                        addr += bytes([0x20])
                addr += encoded_dst_ssid

                for i in range(0,6):
                    if (i < len(self.src_call)):
                        addr += bytes([self.src_call[i]<<1])
                    else:
                        addr += bytes([0x20])
                addr += encoded_src_ssid

                data = addr+bytes([AX25.CTRL_UI])+bytes([AX25.PID_NOLAYER3])+data

                data = data.replace(bytes([0xdb]), bytes([0xdb])+bytes([0xdd]))
                data = data.replace(bytes([0xc0]), bytes([0xdb])+bytes([0xdc]))
                kiss_frame = bytes([KISS.FEND])+bytes([0x00])+data+bytes([KISS.FEND])

                written = self.serial.write(kiss_frame)
                self.txb += datalen

                if written != len(kiss_frame):
                    if self.flow_control:
                        self.interface_ready = True
                    raise IOError("AX.25 interface only wrote "+str(written)+" bytes of "+str(len(kiss_frame)))
            else:
                self.queue(data)

    def queue(self, data):
        self.packet_queue.append(data)

    def process_queue(self):
        if len(self.packet_queue) > 0:
            data = self.packet_queue.pop(0)
            self.interface_ready = True
            self.process_outgoing(data)
        elif len(self.packet_queue) == 0:
            self.interface_ready = True

    def readLoop(self):
        try:
            in_frame = False
            escape = False
            command = KISS.CMD_UNKNOWN
            data_buffer = b""
            last_read_ms = int(time.time()*1000)

            while self.serial.is_open:
                if self.serial.in_waiting:
                    byte = ord(self.serial.read(1))
                    last_read_ms = int(time.time()*1000)

                    if (in_frame and byte == KISS.FEND and command == KISS.CMD_DATA):
                        in_frame = False
                        self.process_incoming(data_buffer)
                    elif (byte == KISS.FEND):
                        in_frame = True
                        command = KISS.CMD_UNKNOWN
                        data_buffer = b""
                    elif (in_frame and len(data_buffer) < self.HW_MTU+AX25.HEADER_SIZE):
                        if (len(data_buffer) == 0 and command == KISS.CMD_UNKNOWN):
                            # We only support one HDLC port for now, so
                            # strip off the port nibble
                            byte = byte & 0x0F
                            command = byte
                        elif (command == KISS.CMD_DATA):
                            if (byte == KISS.FESC):
                                escape = True
                            else:
                                if (escape):
                                    if (byte == KISS.TFEND):
                                        byte = KISS.FEND
                                    if (byte == KISS.TFESC):
                                        byte = KISS.FESC
                                    escape = False
                                data_buffer = data_buffer+bytes([byte])
                        elif (command == KISS.CMD_READY):
                            self.process_queue()
                else:
                    time_since_last = int(time.time()*1000) - last_read_ms
                    if len(data_buffer) > 0 and time_since_last > self.timeout:
                        data_buffer = b""
                        in_frame = False
                        command = KISS.CMD_UNKNOWN
                        escape = False
                    sleep(0.05)

                    if self.flow_control:
                        if not self.interface_ready:
                            if time.time() > self.flow_control_locked + self.flow_control_timeout:
                                RNS.log("Interface "+str(self)+" is unlocking flow control due to time-out. This should not happen. Your hardware might have missed a flow-control READY command, or maybe it does not support flow-control.", RNS.LOG_WARNING)
                                self.process_queue()

        except Exception as e:
            self.online = False
            RNS.log("A serial port error occurred, the contained exception was: "+str(e), RNS.LOG_ERROR)
            RNS.log("The interface "+str(self)+" experienced an unrecoverable error and is now offline.", RNS.LOG_ERROR)
            
            if RNS.Reticulum.panic_on_interface_error:
                RNS.panic()

            RNS.log("Reticulum will attempt to reconnect the interface periodically.", RNS.LOG_ERROR)

        self.online = False
        self.serial.close()
        self.reconnect_port()

    def reconnect_port(self):
        while not self.online:
            try:
                time.sleep(5)
                RNS.log("Attempting to reconnect serial port "+str(self.port)+" for "+str(self)+"...", RNS.LOG_VERBOSE)
                self.open_port()
                if self.serial.is_open:
                    self.configure_device()
            except Exception as e:
                RNS.log("Error while reconnecting port, the contained exception was: "+str(e), RNS.LOG_ERROR)

        RNS.log("Reconnected serial port for "+str(self))

    def should_ingress_limit(self):
        return False

    def __str__(self):
        return "AX25KISSInterface["+self.name+"]"


# Reticulum License
#
# Copyright (c) 2016-2025 Mark Qvist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# - The Software shall not be used in any kind of system which includes amongst
#   its functions the ability to purposefully do harm to human beings.
#
# - The Software shall not be used, directly or indirectly, in the creation of
#   an artificial intelligence, machine learning or language model training
#   dataset, including but not limited to any use that contributes to the
#   training or development of such a model or algorithm.
#
# - The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations
import collections
import enum
import threading
import time
from types import TracebackType
from typing import Type, Callable, TypeVar, Generic, NewType
import abc
import contextlib
import struct
import RNS
from abc import ABC, abstractmethod
TPacket = TypeVar("TPacket")

class SystemMessageTypes(enum.IntEnum):
    SMT_STREAM_DATA = 0xff00

class ChannelOutletBase(ABC, Generic[TPacket]):
    """
    An abstract transport layer interface used by Channel.

    DEPRECATED: This was created for testing; eventually
    Channel will use Link or a LinkBase interface
    directly.
    """
    @abstractmethod
    def send(self, raw: bytes) -> TPacket:
        raise NotImplemented()

    @abstractmethod
    def resend(self, packet: TPacket) -> TPacket:
        raise NotImplemented()

    @property
    @abstractmethod
    def mdu(self):
        raise NotImplemented()

    @property
    @abstractmethod
    def rtt(self):
        raise NotImplemented()

    @property
    @abstractmethod
    def is_usable(self):
        raise NotImplemented()

    @abstractmethod
    def get_packet_state(self, packet: TPacket) -> MessageState:
        raise NotImplemented()

    @abstractmethod
    def timed_out(self):
        raise NotImplemented()

    @abstractmethod
    def __str__(self):
        raise NotImplemented()

    @abstractmethod
    def set_packet_timeout_callback(self, packet: TPacket, callback: Callable[[TPacket], None] | None,
                                    timeout: float | None = None):
        raise NotImplemented()

    @abstractmethod
    def set_packet_delivered_callback(self, packet: TPacket, callback: Callable[[TPacket], None] | None):
        raise NotImplemented()

    @abstractmethod
    def get_packet_id(self, packet: TPacket) -> any:
        raise NotImplemented()


class CEType(enum.IntEnum):
    """
    ChannelException type codes
    """
    ME_NO_MSG_TYPE      = 0
    ME_INVALID_MSG_TYPE = 1
    ME_NOT_REGISTERED   = 2
    ME_LINK_NOT_READY   = 3
    ME_ALREADY_SENT     = 4
    ME_TOO_BIG          = 5


class ChannelException(Exception):
    """
    An exception thrown by Channel, with a type code.
    """
    def __init__(self, ce_type: CEType, *args):
        super().__init__(args)
        self.type = ce_type


class MessageState(enum.IntEnum):
    """
    Set of possible states for a Message
    """
    MSGSTATE_NEW       = 0
    MSGSTATE_SENT      = 1
    MSGSTATE_DELIVERED = 2
    MSGSTATE_FAILED    = 3


class MessageBase(abc.ABC):
    """
    Base type for any messages sent or received on a Channel.
    Subclasses must define the two abstract methods as well as
    the ``MSGTYPE`` class variable.
    """
    # MSGTYPE must be unique within all classes sent over a
    # channel. Additionally, MSGTYPE > 0xf000 are reserved.
    MSGTYPE = None
    """
    Defines a unique identifier for a message class.

    * Must be unique within all classes registered with a ``Channel``
    * Must be less than ``0xf000``. Values greater than or equal to ``0xf000`` are reserved.
    """

    @abstractmethod
    def pack(self) -> bytes:
        """
        Create and return the binary representation of the message

        :return: binary representation of message
        """
        raise NotImplemented()

    @abstractmethod
    def unpack(self, raw: bytes):
        """
        Populate message from binary representation

        :param raw: binary representation
        """
        raise NotImplemented()


MessageCallbackType = NewType("MessageCallbackType", Callable[[MessageBase], bool])


class Envelope:
    """
    Internal wrapper used to transport messages over a channel and
    track its state within the channel framework.
    """
    def unpack(self, message_factories: dict[int, Type]) -> MessageBase:
        msgtype, self.sequence, length = struct.unpack(">HHH", self.raw[:6])
        raw = self.raw[6:]
        ctor = message_factories.get(msgtype, None)
        if ctor is None:
            raise ChannelException(CEType.ME_NOT_REGISTERED, f"Unable to find constructor for Channel MSGTYPE {hex(msgtype)}")
        message = ctor()
        message.unpack(raw)
        self.unpacked = True
        self.message = message

        return message

    def pack(self) -> bytes:
        if self.message.__class__.MSGTYPE is None:
            raise ChannelException(CEType.ME_NO_MSG_TYPE, f"{self.message.__class__} lacks MSGTYPE")
        data = self.message.pack()
        self.raw = struct.pack(">HHH", self.message.MSGTYPE, self.sequence, len(data)) + data
        self.packed = True
        return self.raw

    def __init__(self, outlet: ChannelOutletBase, message: MessageBase = None, raw: bytes = None, sequence: int = None):
        self.ts = time.time()
        self.id = id(self)
        self.message = message
        self.raw = raw
        self.packet: TPacket = None
        self.sequence = sequence
        self.outlet = outlet
        self.tries = 0
        self.unpacked = False
        self.packed = False
        self.tracked = False


class Channel(contextlib.AbstractContextManager):
    """
    Provides reliable delivery of messages over
    a link.

    ``Channel`` differs from ``Request`` and
    ``Resource`` in some important ways:

     **Continuous**
        Messages can be sent or received as long as
        the ``Link`` is open.
     **Bi-directional**
        Messages can be sent in either direction on
        the ``Link``; neither end is the client or
        server.
     **Size-constrained**
        Messages must be encoded into a single packet.

    ``Channel`` is similar to ``Packet``, except that it
    provides reliable delivery (automatic retries) as well
    as a structure for exchanging several types of
    messages over the ``Link``.

    ``Channel`` is not instantiated directly, but rather
    obtained from a ``Link`` with ``get_channel()``.
    """

    # The initial window size at channel setup
    WINDOW     = 2

    # Absolute minimum window size
    WINDOW_MIN = 2
    WINDOW_MIN_LIMIT_SLOW    = 2
    WINDOW_MIN_LIMIT_MEDIUM  = 5
    WINDOW_MIN_LIMIT_FAST    = 16

    # The maximum window size for transfers on slow links
    WINDOW_MAX_SLOW      = 5

    # The maximum window size for transfers on mid-speed links
    WINDOW_MAX_MEDIUM    = 12

    # The maximum window size for transfers on fast links
    WINDOW_MAX_FAST      = 48

    # For calculating maps and guard segments, this
    # must be set to the global maximum window.
    WINDOW_MAX           = WINDOW_MAX_FAST

    # If the fast rate is sustained for this many request
    # rounds, the fast link window size will be allowed.
    FAST_RATE_THRESHOLD  = 10

    # If the RTT rate is higher than this value,
    # the max window size for fast links will be used.
    RTT_FAST            = 0.18
    RTT_MEDIUM          = 0.75
    RTT_SLOW            = 1.45

    # The minimum allowed flexibility of the window size.
    # The difference between window_max and window_min
    # will never be smaller than this value.
    WINDOW_FLEXIBILITY   = 4

    SEQ_MAX     = 0xFFFF
    SEQ_MODULUS = SEQ_MAX+1

    def __init__(self, outlet: ChannelOutletBase):
        """

        @param outlet:
        """
        self._outlet = outlet
        self._lock = threading.RLock()
        self._send_lock = threading.Lock()
        self._tx_ring: collections.deque[Envelope] = collections.deque()
        self._rx_ring: collections.deque[Envelope] = collections.deque()
        self._message_callbacks: [MessageCallbackType] = []
        self._next_sequence = 0
        self._next_rx_sequence = 0
        self._message_factories: dict[int, Type[MessageBase]] = {}
        self._max_tries = 5
        self.fast_rate_rounds    = 0
        self.medium_rate_rounds  = 0

        if self._outlet.rtt > Channel.RTT_SLOW:
            self.window              = 1
            self.window_max          = 1
            self.window_min          = 1
            self.window_flexibility  = 1
        else:
            self.window              = Channel.WINDOW
            self.window_max          = Channel.WINDOW_MAX_SLOW
            self.window_min          = Channel.WINDOW_MIN
            self.window_flexibility  = Channel.WINDOW_FLEXIBILITY

    def __enter__(self) -> Channel:
        return self

    def __exit__(self, __exc_type: Type[BaseException] | None, __exc_value: BaseException | None,
                 __traceback: TracebackType | None) -> bool | None:
        self._shutdown()
        return False

    def register_message_type(self, message_class: Type[MessageBase]):
        """
        Register a message class for reception over a ``Channel``.

        Message classes must extend ``MessageBase``.

        :param message_class: Class to register
        """
        self._register_message_type(message_class, is_system_type=False)

    def _register_message_type(self, message_class: Type[MessageBase], *, is_system_type: bool = False):
        with self._lock:
            if not issubclass(message_class, MessageBase):
                raise ChannelException(CEType.ME_INVALID_MSG_TYPE,
                                       f"{message_class} is not a subclass of {MessageBase}.")
            if message_class.MSGTYPE is None:
                raise ChannelException(CEType.ME_INVALID_MSG_TYPE,
                                       f"{message_class} has invalid MSGTYPE class attribute.")
            if message_class.MSGTYPE >= 0xf000 and not is_system_type:
                raise ChannelException(CEType.ME_INVALID_MSG_TYPE,
                                       f"{message_class} has system-reserved message type.")
            try:
                message_class()
            except Exception as ex:
                raise ChannelException(CEType.ME_INVALID_MSG_TYPE,
                                       f"{message_class} raised an exception when constructed with no arguments: {ex}")

            self._message_factories[message_class.MSGTYPE] = message_class

    def add_message_handler(self, callback: MessageCallbackType):
        """
        Add a handler for incoming messages. A handler
        has the following signature:

        ``(message: MessageBase) -> bool``

        Handlers are processed in the order they are
        added. If any handler returns True, processing
        of the message stops; handlers after the
        returning handler will not be called.

        :param callback: Function to call
        """
        with self._lock:
            if callback not in self._message_callbacks:
                self._message_callbacks.append(callback)

    def remove_message_handler(self, callback: MessageCallbackType):
        """
        Remove a handler added with ``add_message_handler``.

        :param callback: handler to remove
        """
        with self._lock:
            if callback in self._message_callbacks:
                self._message_callbacks.remove(callback)

    def _shutdown(self):
        with self._lock:
            self._message_callbacks.clear()
            self._clear_rings()

    def _clear_rings(self):
        with self._lock:
            for envelope in self._tx_ring:
                if envelope.packet is not None:
                    self._outlet.set_packet_timeout_callback(envelope.packet, None)
                    self._outlet.set_packet_delivered_callback(envelope.packet, None)
                envelope.tracked = False
            for envelope in self._rx_ring:
                envelope.tracked = False
            self._tx_ring.clear()
            self._rx_ring.clear()

    def _emplace_envelope(self, envelope: Envelope, ring: collections.deque[Envelope]) -> bool:
        with self._lock:
            i = 0

            for existing in ring:

                if envelope.sequence == existing.sequence:
                    RNS.log(f"Envelope: Emplacement of duplicate envelope with sequence "+str(envelope.sequence), RNS.LOG_EXTREME)
                    return False

                if envelope.sequence < existing.sequence and not (self._next_rx_sequence - envelope.sequence) > (Channel.SEQ_MAX//2):
                    ring.insert(i, envelope)

                    envelope.tracked = True
                    return True

                i += 1

            envelope.tracked = True
            ring.append(envelope)

            return True

    def _run_callbacks(self, message: MessageBase):
        cbs = self._message_callbacks.copy()

        for cb in cbs:
            try:
                if cb(message):
                    return
            except Exception as e:
                RNS.log("Channel "+str(self)+" experienced an error while running a message callback. The contained exception was: "+str(e), RNS.LOG_ERROR)

    def _receive(self, raw: bytes):
        try:
            envelope = Envelope(outlet=self._outlet, raw=raw)
            with self._lock:
                message = envelope.unpack(self._message_factories)

                if envelope.sequence < self._next_rx_sequence:
                    window_overflow = (self._next_rx_sequence+Channel.WINDOW_MAX) % Channel.SEQ_MODULUS
                    if window_overflow < self._next_rx_sequence:
                        if envelope.sequence > window_overflow:
                            RNS.log("Invalid packet sequence ("+str(envelope.sequence)+") received on channel "+str(self), RNS.LOG_EXTREME)
                            return
                    else:
                        RNS.log("Invalid packet sequence ("+str(envelope.sequence)+") received on channel "+str(self), RNS.LOG_EXTREME)
                        return

                is_new = self._emplace_envelope(envelope, self._rx_ring)

            if not is_new:
                RNS.log("Duplicate message received on channel "+str(self), RNS.LOG_EXTREME)
                return
            else:
                with self._lock:
                    contigous = []
                    for e in self._rx_ring:
                        if e.sequence == self._next_rx_sequence:
                            contigous.append(e)
                            self._next_rx_sequence = (self._next_rx_sequence + 1) % Channel.SEQ_MODULUS
                            if self._next_rx_sequence == 0:
                                for e in self._rx_ring:
                                    if e.sequence == self._next_rx_sequence:
                                        contigous.append(e)
                                        self._next_rx_sequence = (self._next_rx_sequence + 1) % Channel.SEQ_MODULUS

                    for e in contigous:
                        if not e.unpacked:
                            m = e.unpack(self._message_factories)
                        else:
                            m = e.message

                        self._rx_ring.remove(e)
                        self._run_callbacks(m)

        except Exception as e:
            RNS.log("An error ocurred while receiving data on "+str(self)+". The contained exception was: "+str(e), RNS.LOG_ERROR)

    def is_ready_to_send(self) -> bool:
        """
        Check if ``Channel`` is ready to send.

        :return: True if ready
        """
        if not self._outlet.is_usable:
            return False

        with self._lock:
            outstanding = 0
            for envelope in self._tx_ring:
                if envelope.outlet == self._outlet:
                    if not envelope.packet or not self._outlet.get_packet_state(envelope.packet) == MessageState.MSGSTATE_DELIVERED:
                        outstanding += 1

            if outstanding >= self.window:
                return False

        return True

    def _packet_tx_op(self, packet: TPacket, op: Callable[[TPacket], bool]):
        target_id = self._outlet.get_packet_id(packet)
        with self._lock:
            envelope = next(filter(lambda e: e.packet is not None
                                             and self._outlet.get_packet_id(e.packet) == target_id,
                                   self._tx_ring), None)

            if envelope and op(envelope):
                envelope.tracked = False
                if envelope in self._tx_ring:
                    self._tx_ring.remove(envelope)

                    if self.window < self.window_max:
                        self.window += 1

                        # TODO: Remove at some point
                        # RNS.log("Increased "+str(self)+" window to "+str(self.window), RNS.LOG_DEBUG)

                    if self._outlet.rtt != 0:
                        if self._outlet.rtt > Channel.RTT_FAST:
                            self.fast_rate_rounds = 0

                            if self._outlet.rtt > Channel.RTT_MEDIUM:
                                self.medium_rate_rounds = 0

                            else:
                                self.medium_rate_rounds += 1
                                if self.window_max < Channel.WINDOW_MAX_MEDIUM and self.medium_rate_rounds == Channel.FAST_RATE_THRESHOLD:
                                    self.window_max = Channel.WINDOW_MAX_MEDIUM
                                    self.window_min = Channel.WINDOW_MIN_LIMIT_MEDIUM
                                    # TODO: Remove at some point
                                    # RNS.log("Increased "+str(self)+" max window to "+str(self.window_max), RNS.LOG_DEBUG)
                                    # RNS.log("Increased "+str(self)+" min window to "+str(self.window_min), RNS.LOG_DEBUG)

                        else:
                            self.fast_rate_rounds += 1
                            if self.window_max < Channel.WINDOW_MAX_FAST and self.fast_rate_rounds == Channel.FAST_RATE_THRESHOLD:
                                self.window_max = Channel.WINDOW_MAX_FAST
                                self.window_min = Channel.WINDOW_MIN_LIMIT_FAST
                                # TODO: Remove at some point
                                # RNS.log("Increased "+str(self)+" max window to "+str(self.window_max), RNS.LOG_DEBUG)
                                # RNS.log("Increased "+str(self)+" min window to "+str(self.window_min), RNS.LOG_DEBUG)


                else:
                    RNS.log("Envelope not found in TX ring for "+str(self), RNS.LOG_EXTREME)
        if not envelope:
            RNS.log("Spurious message received on "+str(self), RNS.LOG_EXTREME)

    def _packet_delivered(self, packet: TPacket):
        self._packet_tx_op(packet, lambda env: True)

    def _update_packet_timeouts(self):
        for envelope in self._tx_ring:
            updated_timeout = self._get_packet_timeout_time(envelope.tries)
            if envelope.packet and hasattr(envelope.packet, "receipt") and envelope.packet.receipt and envelope.packet.receipt.timeout:
                if updated_timeout > envelope.packet.receipt.timeout:
                    envelope.packet.receipt.set_timeout(updated_timeout)

    def _get_packet_timeout_time(self, tries: int) -> float:
        to = pow(1.5, tries - 1) * max(self._outlet.rtt*2.5, 0.025) * (len(self._tx_ring)+1.5)
        return to

    def _packet_timeout(self, packet: TPacket):
        if self._outlet.get_packet_state(packet) == MessageState.MSGSTATE_DELIVERED:
            return

        target_id = self._outlet.get_packet_id(packet)
        envelope_to_resend: Envelope | None = None
        should_teardown = False
        with self._lock:
            envelope = next(filter(
                lambda e: e.packet is not None and self._outlet.get_packet_id(e.packet) == target_id,
                self._tx_ring), None)
            if envelope is None:
                return

            if envelope.tries >= self._max_tries:
                should_teardown = True
            else:
                envelope.tries += 1
                envelope_to_resend = envelope

                if self.window > self.window_min:
                    self.window -= 1
                    if self.window_max > (self.window_min+self.window_flexibility):
                        self.window_max -= 1

        if should_teardown:
            RNS.log("Retry count exceeded on "+str(self)+", tearing down Link.", RNS.LOG_ERROR)
            self._shutdown()
            self._outlet.timed_out()
            return

        if envelope_to_resend is not None:
            self._outlet.resend(envelope_to_resend.packet)
            with self._lock:
                self._outlet.set_packet_delivered_callback(envelope_to_resend.packet, self._packet_delivered)
                self._outlet.set_packet_timeout_callback(
                    envelope_to_resend.packet, self._packet_timeout,
                    self._get_packet_timeout_time(envelope_to_resend.tries))
                self._update_packet_timeouts()
                already_delivered = (self._outlet.get_packet_state(envelope_to_resend.packet) == MessageState.MSGSTATE_DELIVERED)

            if already_delivered:
                self._packet_delivered(envelope_to_resend.packet)

    def send(self, message: MessageBase) -> Envelope:
        """
        Send a message. If a message send is attempted and
        ``Channel`` is not ready, an exception is thrown.

        :param message: an instance of a ``MessageBase`` subclass
        """
        with self._send_lock:
            with self._lock:
                if not self.is_ready_to_send():
                    raise ChannelException(CEType.ME_LINK_NOT_READY, f"Link is not ready")

                reserved_sequence = self._next_sequence
                envelope = Envelope(self._outlet, message=message, sequence=reserved_sequence)
                envelope.pack()
                if len(envelope.raw) > self._outlet.mdu:
                    raise ChannelException(CEType.ME_TOO_BIG,
                                           f"Packed message too big for packet: {len(envelope.raw)} > {self._outlet.mdu}")
                self._next_sequence = (reserved_sequence + 1) % Channel.SEQ_MODULUS

            envelope.packet = self._outlet.send(envelope.raw)

            if (envelope.packet is None
                    or getattr(envelope.packet, "raw", None) is None
                    or (hasattr(envelope.packet, "receipt") and envelope.packet.receipt is None)):
                with self._lock:
                    self._next_sequence = reserved_sequence
                raise ChannelException(CEType.ME_LINK_NOT_READY, "Outlet did not transmit packet")

            with self._lock:
                self._emplace_envelope(envelope, self._tx_ring)
                envelope.tries += 1
                self._outlet.set_packet_delivered_callback(envelope.packet, self._packet_delivered)
                self._outlet.set_packet_timeout_callback(envelope.packet, self._packet_timeout, self._get_packet_timeout_time(envelope.tries))
                self._update_packet_timeouts()
                already_delivered = (self._outlet.get_packet_state(envelope.packet) == MessageState.MSGSTATE_DELIVERED)

        # prevent _tx_ring envelope leak
        if already_delivered:
            self._packet_delivered(envelope.packet)

        return envelope

    @property
    def mdu(self):
        """
        Maximum Data Unit: the number of bytes available
        for a message to consume in a single send. This
        value is adjusted from the ``Link`` MDU to accommodate
        message header information.

        :return: number of bytes available
        """
        mdu = self._outlet.mdu - 6  # sizeof(msgtype) + sizeof(length) + sizeof(sequence)
        if mdu > 0xFFFF:
            mdu = 0xFFFF
        return mdu


class LinkChannelOutlet(ChannelOutletBase):
    """
    An implementation of ChannelOutletBase for RNS.Link.
    Allows Channel to send packets over an RNS Link with
    Packets.

    :param link: RNS Link to wrap
    """
    def __init__(self, link: RNS.Link):
        self.link = link

    def send(self, raw: bytes) -> RNS.Packet:
        packet = RNS.Packet(self.link, raw, context=RNS.Packet.CHANNEL)
        if self.link.status == RNS.Link.ACTIVE:
            packet.send()
        return packet

    def resend(self, packet: RNS.Packet) -> RNS.Packet:
        receipt = packet.resend()
        if not receipt:
            RNS.log("Failed to resend packet", RNS.LOG_ERROR)
        return packet

    @property
    def mdu(self):
        return self.link.mdu

    @property
    def rtt(self):
        return self.link.rtt

    @property
    def is_usable(self):
        return True  # had issues looking at Link.status

    def get_packet_state(self, packet: TPacket) -> MessageState:
        if packet.receipt == None:
            return MessageState.MSGSTATE_FAILED

        status = packet.receipt.get_status()
        if status == RNS.PacketReceipt.SENT:
            return MessageState.MSGSTATE_SENT
        if status == RNS.PacketReceipt.DELIVERED:
            return MessageState.MSGSTATE_DELIVERED
        if status == RNS.PacketReceipt.FAILED:
            return MessageState.MSGSTATE_FAILED
        else:
            raise Exception(f"Unexpected receipt state: {status}")

    def timed_out(self):
        self.link.teardown()

    def __str__(self):
        return f"{self.__class__.__name__}({self.link})"

    def set_packet_timeout_callback(self, packet: RNS.Packet, callback: Callable[[RNS.Packet], None] | None,
                                    timeout: float | None = None):
        if timeout and packet.receipt:
            packet.receipt.set_timeout(timeout)

        def inner(receipt: RNS.PacketReceipt):
            callback(packet)

        if packet and packet.receipt:
            packet.receipt.set_timeout_callback(inner if callback else None)

    def set_packet_delivered_callback(self, packet: RNS.Packet, callback: Callable[[RNS.Packet], None] | None):
        def inner(receipt: RNS.PacketReceipt):
            callback(packet)

        if packet and packet.receipt:
            packet.receipt.set_delivery_callback(inner if callback else None)

    def get_packet_id(self, packet: RNS.Packet) -> any:
        if (packet
                and getattr(packet, "raw", None) is not None
                and hasattr(packet, "get_hash")
                and callable(packet.get_hash)):
            return packet.get_hash()
        else:
            return None



# validate.py
# -*- coding: utf-8 -*-
# pylint: disable=
#
# A Validator object.
#
# Copyright (C) 2005-2014:
# (name) : (email)
# Michael Foord: fuzzyman AT voidspace DOT org DOT uk
# Mark Andrews: mark AT la-la DOT com
# Nicola Larosa: nico AT tekNico DOT net
# Rob Dennis: rdennis AT gmail DOT com
# Eli Courtwright: eli AT courtwright DOT org

# This software is licensed under the terms of the BSD license.
# http://opensource.org/licenses/BSD-3-Clause

# ConfigObj 5 - main repository for documentation and issue tracking:
# https://github.com/DiffSK/configobj

import re
import sys
from pprint import pprint

__version__ = '1.0.1'

__all__ = (
    'dottedQuadToNum',
    'numToDottedQuad',
    'ValidateError',
    'VdtUnknownCheckError',
    'VdtParamError',
    'VdtTypeError',
    'VdtValueError',
    'VdtValueTooSmallError',
    'VdtValueTooBigError',
    'VdtValueTooShortError',
    'VdtValueTooLongError',
    'VdtMissingValue',
    'Validator',
    'is_integer',
    'is_float',
    'is_boolean',
    'is_list',
    'is_tuple',
    'is_ip_addr',
    'is_string',
    'is_int_list',
    'is_bool_list',
    'is_float_list',
    'is_string_list',
    'is_ip_addr_list',
    'is_mixed_list',
    'is_option',
)

_list_arg = re.compile(r'''
    (?:
        ([a-zA-Z_][a-zA-Z0-9_]*)\s*=\s*list\(
            (
                (?:
                    \s*
                    (?:
                        (?:".*?")|              # double quotes
                        (?:'.*?')|              # single quotes
                        (?:[^'",\s\)][^,\)]*?)  # unquoted
                    )
                    \s*,\s*
                )*
                (?:
                    (?:".*?")|              # double quotes
                    (?:'.*?')|              # single quotes
                    (?:[^'",\s\)][^,\)]*?)  # unquoted
                )?                          # last one
            )
        \)
    )
''', re.VERBOSE | re.DOTALL)    # two groups

_list_members = re.compile(r'''
    (
        (?:".*?")|              # double quotes
        (?:'.*?')|              # single quotes
        (?:[^'",\s=][^,=]*?)       # unquoted
    )
    (?:
    (?:\s*,\s*)|(?:\s*$)            # comma
    )
''', re.VERBOSE | re.DOTALL)    # one group

_paramstring = r'''
    (?:
        (
            (?:
                [a-zA-Z_][a-zA-Z0-9_]*\s*=\s*list\(
                    (?:
                        \s*
                        (?:
                            (?:".*?")|              # double quotes
                            (?:'.*?')|              # single quotes
                            (?:[^'",\s\)][^,\)]*?)       # unquoted
                        )
                        \s*,\s*
                    )*
                    (?:
                        (?:".*?")|              # double quotes
                        (?:'.*?')|              # single quotes
                        (?:[^'",\s\)][^,\)]*?)       # unquoted
                    )?                              # last one
                \)
            )|
            (?:
                (?:".*?")|              # double quotes
                (?:'.*?')|              # single quotes
                (?:[^'",\s=][^,=]*?)|       # unquoted
                (?:                         # keyword argument
                    [a-zA-Z_][a-zA-Z0-9_]*\s*=\s*
                    (?:
                        (?:".*?")|              # double quotes
                        (?:'.*?')|              # single quotes
                        (?:[^'",\s=][^,=]*?)       # unquoted
                    )
                )
            )
        )
        (?:
            (?:\s*,\s*)|(?:\s*$)            # comma
        )
    )
    '''

_matchstring = '^%s*' % _paramstring

def dottedQuadToNum(ip):
    # import here to avoid it when ip_addr values are not used
    import socket, struct

    try:
        return struct.unpack('!L',
            socket.inet_aton(ip.strip()))[0]
    except socket.error:
        raise ValueError('Not a good dotted-quad IP: %s' % ip)
    return

def numToDottedQuad(num):
    # import here to avoid it when ip_addr values are not used
    import socket, struct

    # no need to intercept here, 4294967295L is fine
    if num > int(4294967295) or num < 0:
        raise ValueError('Not a good numeric IP: %s' % num)
    try:
        return socket.inet_ntoa(
            struct.pack('!L', int(num)))
    except (socket.error, struct.error, OverflowError):
        raise ValueError('Not a good numeric IP: %s' % num)

class ValidateError(Exception):
    """
    This error indicates that the check failed.
    It can be the base class for more specific errors.
    """

class VdtMissingValue(ValidateError):
    """No value was supplied to a check that needed one."""

class VdtUnknownCheckError(ValidateError):
    def __init__(self, value):
        ValidateError.__init__(self, 'the check "{}" is unknown.'.format(value))


class VdtParamError(SyntaxError):
    NOT_GIVEN = object()

    def __init__(self, name_or_msg, value=NOT_GIVEN):
        if value is self.NOT_GIVEN:
            SyntaxError.__init__(self, name_or_msg)
        else:
            SyntaxError.__init__(self, 'passed an incorrect value "{}" for parameter "{}".'.format(value, name_or_msg))


class VdtTypeError(ValidateError):
    def __init__(self, value):
        ValidateError.__init__(self, 'the value "{}" is of the wrong type.'.format(value))


class VdtValueError(ValidateError):
    def __init__(self, value):
        ValidateError.__init__(self, 'the value "{}" is unacceptable.'.format(value))


class VdtValueTooSmallError(VdtValueError):
    def __init__(self, value):
        ValidateError.__init__(self, 'the value "{}" is too small.'.format(value))


class VdtValueTooBigError(VdtValueError):
    def __init__(self, value):
        ValidateError.__init__(self, 'the value "{}" is too big.'.format(value))


class VdtValueTooShortError(VdtValueError):
    def __init__(self, value):
        ValidateError.__init__(
            self,
            'the value "{}" is too short.'.format(value))

class VdtValueTooLongError(VdtValueError):
    def __init__(self, value):
        ValidateError.__init__(self, 'the value "{}" is too long.'.format(value))

class Validator(object):
    # this regex does the initial parsing of the checks
    _func_re = re.compile(r'([^\(\)]+?)\((.*)\)', re.DOTALL)

    # this regex takes apart keyword arguments
    _key_arg = re.compile(r'^([a-zA-Z_][a-zA-Z0-9_]*)\s*=\s*(.*)$',  re.DOTALL)


    # this regex finds keyword=list(....) type values
    _list_arg = _list_arg

    # this regex takes individual values out of lists - in one pass
    _list_members = _list_members

    # These regexes check a set of arguments for validity
    # and then pull the members out
    _paramfinder = re.compile(_paramstring, re.VERBOSE | re.DOTALL)
    _matchfinder = re.compile(_matchstring, re.VERBOSE | re.DOTALL)

    def __init__(self, functions=None):
        self.functions = {
            '': self._pass,
            'integer': is_integer,
            'float': is_float,
            'boolean': is_boolean,
            'ip_addr': is_ip_addr,
            'string': is_string,
            'list': is_list,
            'tuple': is_tuple,
            'int_list': is_int_list,
            'float_list': is_float_list,
            'bool_list': is_bool_list,
            'ip_addr_list': is_ip_addr_list,
            'string_list': is_string_list,
            'mixed_list': is_mixed_list,
            'pass': self._pass,
            'option': is_option,
            'force_list': force_list,
        }
        if functions is not None:
            self.functions.update(functions)
        # tekNico: for use by ConfigObj
        self.baseErrorClass = ValidateError
        self._cache = {}

    def check(self, check, value, missing=False):
        fun_name, fun_args, fun_kwargs, default = self._parse_with_caching(check)

        if missing:
            if default is None:
                # no information needed here - to be handled by caller
                raise VdtMissingValue()
            value = self._handle_none(default)

        if value is None:
            return None

        return self._check_value(value, fun_name, fun_args, fun_kwargs)

    def _handle_none(self, value):
        if value == 'None':
            return None
        elif value in ("'None'", '"None"'):
            # Special case a quoted None
            value = self._unquote(value)
        return value

    def _parse_with_caching(self, check):
        if check in self._cache:
            fun_name, fun_args, fun_kwargs, default = self._cache[check]
            # We call list and dict below to work with *copies* of the data
            # rather than the original (which are mutable of course)
            fun_args = list(fun_args)
            fun_kwargs = dict(fun_kwargs)
        else:
            fun_name, fun_args, fun_kwargs, default = self._parse_check(check)
            fun_kwargs = {str(key): value for (key, value) in list(fun_kwargs.items())}
            self._cache[check] = fun_name, list(fun_args), dict(fun_kwargs), default
        return fun_name, fun_args, fun_kwargs, default

    def _check_value(self, value, fun_name, fun_args, fun_kwargs):
        try:
            fun = self.functions[fun_name]
        except KeyError:
            raise VdtUnknownCheckError(fun_name)
        else:
            return fun(value, *fun_args, **fun_kwargs)

    def _parse_check(self, check):
        fun_match = self._func_re.match(check)
        if fun_match:
            fun_name = fun_match.group(1)
            arg_string = fun_match.group(2)
            arg_match = self._matchfinder.match(arg_string)
            if arg_match is None:
                # Bad syntax
                raise VdtParamError('Bad syntax in check "%s".' % check)
            fun_args = []
            fun_kwargs = {}
            # pull out args of group 2
            for arg in self._paramfinder.findall(arg_string):
                # args may need whitespace removing (before removing quotes)
                arg = arg.strip()
                listmatch = self._list_arg.match(arg)
                if listmatch:
                    key, val = self._list_handle(listmatch)
                    fun_kwargs[key] = val
                    continue
                keymatch = self._key_arg.match(arg)
                if keymatch:
                    val = keymatch.group(2)
                    if not val in ("'None'", '"None"'):
                        # Special case a quoted None
                        val = self._unquote(val)
                    fun_kwargs[keymatch.group(1)] = val
                    continue

                fun_args.append(self._unquote(arg))
        else:
            # allows for function names without (args)
            return check, (), {}, None

        # Default must be deleted if the value is specified too,
        # otherwise the check function will get a spurious "default" keyword arg
        default = fun_kwargs.pop('default', None)
        return fun_name, fun_args, fun_kwargs, default

    def _unquote(self, val):
        if (len(val) >= 2) and (val[0] in ("'", '"')) and (val[0] == val[-1]):
            val = val[1:-1]
        return val

    def _list_handle(self, listmatch):
        out = []
        name = listmatch.group(1)
        args = listmatch.group(2)
        for arg in self._list_members.findall(args):
            out.append(self._unquote(arg))
        return name, out

    def _pass(self, value):
        return value

    def get_default_value(self, check):
        fun_name, fun_args, fun_kwargs, default = self._parse_with_caching(check)
        if default is None:
            raise KeyError('Check "%s" has no default value.' % check)
        value = self._handle_none(default)
        if value is None:
            return value
        return self._check_value(value, fun_name, fun_args, fun_kwargs)

def _is_num_param(names, values, to_float=False):
    fun = to_float and float or int
    out_params = []
    for (name, val) in zip(names, values):
        if val is None:
            out_params.append(val)
        elif isinstance(val, (int, float, str)):
            try:
                out_params.append(fun(val))
            except ValueError:
                raise VdtParamError(name, val)
        else:
            raise VdtParamError(name, val)
    return out_params

# built in checks
# you can override these by setting the appropriate name
# in Validator.functions
# note: if the params are specified wrongly in your input string,
#       you will also raise errors.
def is_integer(value, min=None, max=None):
    (min_val, max_val) = _is_num_param(  # pylint: disable=unbalanced-tuple-unpacking
        ('min', 'max'), (min, max))
    if not isinstance(value, (int, str)):
        raise VdtTypeError(value)
    if isinstance(value, str):
        # if it's a string - does it represent an integer ?
        try:
            value = int(value)
        except ValueError:
            raise VdtTypeError(value)
    if (min_val is not None) and (value < min_val):
        raise VdtValueTooSmallError(value)
    if (max_val is not None) and (value > max_val):
        raise VdtValueTooBigError(value)
    return value


def is_float(value, min=None, max=None):
    (min_val, max_val) = _is_num_param(
        ('min', 'max'), (min, max), to_float=True)
    if not isinstance(value, (int, float, str)):
        raise VdtTypeError(value)
    if not isinstance(value, float):
        # if it's a string - does it represent a float ?
        try:
            value = float(value)
        except ValueError:
            raise VdtTypeError(value)
    if (min_val is not None) and (value < min_val):
        raise VdtValueTooSmallError(value)
    if (max_val is not None) and (value > max_val):
        raise VdtValueTooBigError(value)
    return value

bool_dict = {
    True: True, 'on': True, '1': True, 'true': True, 'yes': True,
    False: False, 'off': False, '0': False, 'false': False, 'no': False,
}

def is_boolean(value):
    if isinstance(value, str):
        try:
            return bool_dict[value.lower()]
        except KeyError:
            raise VdtTypeError(value)
    # we do an equality test rather than an identity test
    # this ensures Python 2.2 compatibility
    # and allows 0 and 1 to represent True and False
    if value == False:
        return False
    elif value == True:
        return True
    else:
        raise VdtTypeError(value)


def is_ip_addr(value):
    if not isinstance(value, str):
        raise VdtTypeError(value)
    value = value.strip()
    try:
        dottedQuadToNum(value)
    except ValueError:
        raise VdtValueError(value)
    return value


def is_list(value, min=None, max=None):
    (min_len, max_len) = _is_num_param(  # pylint: disable=unbalanced-tuple-unpacking
        ('min', 'max'), (min, max))
    if isinstance(value, str):
        raise VdtTypeError(value)
    try:
        num_members = len(value)
    except TypeError:
        raise VdtTypeError(value)
    if min_len is not None and num_members < min_len:
        raise VdtValueTooShortError(value)
    if max_len is not None and num_members > max_len:
        raise VdtValueTooLongError(value)
    return list(value)


def is_tuple(value, min=None, max=None):
    return tuple(is_list(value, min, max))

def is_string(value, min=None, max=None):
    if not isinstance(value, str):
        raise VdtTypeError(value)
    (min_len, max_len) = _is_num_param(
        ('min', 'max'), (min, max))
    try:
        num_members = len(value)
    except TypeError:
        raise VdtTypeError(value)
    if min_len is not None and num_members < min_len:
        raise VdtValueTooShortError(value)
    if max_len is not None and num_members > max_len:
        raise VdtValueTooLongError(value)
    return value


def is_int_list(value, min=None, max=None):
    return [is_integer(mem) for mem in is_list(value, min, max)]

def is_bool_list(value, min=None, max=None):
    return [is_boolean(mem) for mem in is_list(value, min, max)]

def is_float_list(value, min=None, max=None):
    return [is_float(mem) for mem in is_list(value, min, max)]

def is_string_list(value, min=None, max=None):
    if isinstance(value, str):
        raise VdtTypeError(value)
    return [is_string(mem) for mem in is_list(value, min, max)]

def is_ip_addr_list(value, min=None, max=None):
    return [is_ip_addr(mem) for mem in is_list(value, min, max)]

def force_list(value, min=None, max=None):
    if not isinstance(value, (list, tuple)):
        value = [value]
    return is_list(value, min, max)

fun_dict = {
    int: is_integer,
    'int': is_integer,
    'integer': is_integer,
    float: is_float,
    'float': is_float,
    'ip_addr': is_ip_addr,
    str: is_string,
    'str': is_string,
    'string': is_string,
    bool: is_boolean,
    'bool': is_boolean,
    'boolean': is_boolean,
}

def is_mixed_list(value, *args):
    try: length = len(value)
    except TypeError: raise VdtTypeError(value)
    if length < len(args): raise VdtValueTooShortError(value)
    elif length > len(args): raise VdtValueTooLongError(value)
    try: return [fun_dict[arg](val) for arg, val in zip(args, value)]
    except KeyError as cause: raise VdtParamError('mixed_list', cause)


def is_option(value, *options):
    if not isinstance(value, str): raise VdtTypeError(value)
    if not value in options: raise VdtValueError(value)
    return value




# Reticulum License
#
# Copyright (c) 2016-2025 Mark Qvist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# - The Software shall not be used in any kind of system which includes amongst
#   its functions the ability to purposefully do harm to human beings.
#
# - The Software shall not be used, directly or indirectly, in the creation of
#   an artificial intelligence, machine learning or language model training
#   dataset, including but not limited to any use that contributes to the
#   training or development of such a model or algorithm.
#
# - The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import RNS
import threading
import time

from collections import deque
from RNS.Interfaces.Interface import Interface

class HDLC():
    FLAG              = 0x7E
    ESC               = 0x7D
    ESC_MASK          = 0x20

    @staticmethod
    def escape(data):
        data = data.replace(bytes([HDLC.ESC]), bytes([HDLC.ESC, HDLC.ESC^HDLC.ESC_MASK]))
        data = data.replace(bytes([HDLC.FLAG]), bytes([HDLC.ESC, HDLC.FLAG^HDLC.ESC_MASK]))
        return data

class WDCL():
    WDCL_T_DISCOVER        = 0x00
    WDCL_T_CONNECT         = 0x01
    WDCL_T_CMD             = 0x02
    WDCL_T_LOG             = 0x03
    WDCL_T_DISP            = 0x04
    WDCL_T_ENDPOINT_PKT    = 0x05
    WDCL_T_ENCAP_PROTO     = 0x06

    WDCL_BROADCAST         = bytes([0xFF, 0xFF, 0xFF, 0xFF])

    WDCL_HANDSHAKE_TIMEOUT = 2

    HEADER_MINSIZE         = 4+1
    MAX_CHUNK              = 32768
    port                   = None
    speed                  = None
    databits               = None
    parity                 = None
    stopbits               = None
    serial                 = None

    def __init__(self, owner, device, port, as_interface=False):
        import importlib.util
        if RNS.vendor.platformutils.is_android():
            self.on_android  = True
            if importlib.util.find_spec('usbserial4a') != None:
                from usbserial4a import serial4a as serial
                parity = "N"

                if importlib.util.find_spec('jnius') == None:
                    RNS.log("Could not load jnius API wrapper for Android, RNode interface cannot be created.", RNS.LOG_CRITICAL)
                    RNS.log("This probably means you are trying to use an USB-based interface from within Termux or similar.", RNS.LOG_CRITICAL)
                    RNS.log("This is currently not possible, due to this environment limiting access to the native Android APIs.", RNS.LOG_CRITICAL)
                    RNS.panic()

            else:
                RNS.log("Could not load USB serial module for Android, Weave interface cannot be created.", RNS.LOG_CRITICAL)
                RNS.panic()
        
        else:
            self.on_android = False
            if importlib.util.find_spec('serial') != None:
                import serial
                parity = serial.PARITY_NONE
            else:
                RNS.log("Using the Weave interface requires a serial communication module to be installed.", RNS.LOG_CRITICAL)
                RNS.log("You can install one with the command: python3 -m pip install pyserial", RNS.LOG_CRITICAL)
                RNS.panic()

        if not RNS.vendor.platformutils.is_android():
            if port == None: raise ValueError("No port specified")
        
        self.supports_discovery   = True
        self.discovery_frequency  = None
        self.discovery_bandwidth  = None
        self.discovery_channel    = None
        self.discovery_modulation = None

        self.switch_identity = owner.switch_identity
        self.switch_id = self.switch_identity.sig_pub_bytes[-4:]
        self.switch_pub_bytes = self.switch_identity.sig_pub_bytes

        self.rxb = 0
        self.txb = 0
        self.owner = owner
        self.as_interface = as_interface
        self.device = device
        self.device.connection = self
        self.pyserial = serial
        self.serial   = None
        self.port     = port
        self.speed    = 3000000
        self.databits = 8
        self.parity   = parity
        self.stopbits = 1
        self.timeout  = 100
        self.online   = False
        self.frame_buffer = b""
        self.next_tx = 0
        self.should_run = True
        self.receiver = None
        self.wdcl_connected = False
        self.reconnecting = False
        self.frame_queue = deque()
        if not self.as_interface:
            self.id = RNS.Identity.full_hash(port.hwid.encode("utf-8"))

        if self.as_interface:
            try:
                self.open_port()
                if self.serial and self.serial.is_open: self.configure_device()
                else: raise IOError("Could not open serial port")

            except Exception as e:
                RNS.log("Could not open serial port for interface "+str(self), RNS.LOG_ERROR)
                RNS.log("The contained exception was: "+str(e), RNS.LOG_ERROR)
                RNS.log("Reticulum will attempt to bring up this interface periodically", RNS.LOG_ERROR)
                if not self.owner.detached and not self.reconnecting:
                    thread = threading.Thread(target=self.reconnect_port)
                    thread.daemon = True
                    thread.start()

        else:
            try: self.open_port()
            except Exception as e:
                self.owner.wlog("Could not open serial port")
                raise e

            if self.serial.is_open: self.configure_device()
            else: raise IOError("Could not open serial port")


    def open_port(self):
        if not self.on_android:
            if self.as_interface:
                RNS.log(f"Opening serial port {self.port}...", RNS.LOG_VERBOSE)
                target_port = self.port
            else:
                self.owner.wlog(f"Opening serial port {self.port.device}...")
                target_port = self.port.device

            self.serial = self.pyserial.Serial(
                port = target_port,
                baudrate = self.speed,
                bytesize = self.databits,
                parity = self.parity,
                stopbits = self.stopbits,
                xonxoff = False,
                rtscts = False,
                timeout = 0.250,
                inter_byte_timeout = None,
                write_timeout = None,
                dsrdtr = False)

        else:
            if self.port != None:
                # Get device parameters
                from usb4a import usb
                device = usb.get_usb_device(self.port)
                if device:
                    vid = device.getVendorId()
                    pid = device.getProductId()

                    # Driver overrides for speficic chips
                    proxy = self.pyserial.get_serial_port
                    if vid == 0x1A86 and pid == 0x55D4:
                        # Force CDC driver for Qinheng CH34x
                        RNS.log(str(self)+" using CDC driver for "+RNS.hexrep(vid)+":"+RNS.hexrep(pid), RNS.LOG_DEBUG)
                        from usbserial4a.cdcacmserial4a import CdcAcmSerial
                        proxy = CdcAcmSerial

                    self.serial = proxy(
                        self.port,
                        baudrate = self.speed,
                        bytesize = self.databits,
                        parity = self.parity,
                        stopbits = self.stopbits,
                        xonxoff = False,
                        rtscts = False,
                        timeout = None,
                        inter_byte_timeout = None,
                        # write_timeout = wtimeout,
                        dsrdtr = False,
                    )

                    if vid == 0x0403:
                        # Hardware parameters for FTDI devices @ 115200 baud
                        self.serial.DEFAULT_READ_BUFFER_SIZE = 16 * 1024
                        self.serial.USB_READ_TIMEOUT_MILLIS = 100
                        self.serial.timeout = 0.1
                    elif vid == 0x10C4:
                        # Hardware parameters for SiLabs CP210x @ 115200 baud
                        self.serial.DEFAULT_READ_BUFFER_SIZE = 64 
                        self.serial.USB_READ_TIMEOUT_MILLIS = 12
                        self.serial.timeout = 0.012
                    elif vid == 0x1A86 and pid == 0x55D4:
                        # Hardware parameters for Qinheng CH34x @ 115200 baud
                        self.serial.DEFAULT_READ_BUFFER_SIZE = 64
                        self.serial.USB_READ_TIMEOUT_MILLIS = 12
                        self.serial.timeout = 0.1
                    else:
                        # Default values
                        self.serial.DEFAULT_READ_BUFFER_SIZE = 1 * 1024
                        self.serial.USB_READ_TIMEOUT_MILLIS = 100
                        self.serial.timeout = 0.1

                    RNS.log(str(self)+" USB read buffer size set to "+RNS.prettysize(self.serial.DEFAULT_READ_BUFFER_SIZE), RNS.LOG_DEBUG)
                    RNS.log(str(self)+" USB read timeout set to "+str(self.serial.USB_READ_TIMEOUT_MILLIS)+"ms", RNS.LOG_DEBUG)
                    RNS.log(str(self)+" USB write timeout set to "+str(self.serial.USB_WRITE_TIMEOUT_MILLIS)+"ms", RNS.LOG_DEBUG)

    def close(self):
        self.should_run = False
        self.online = False
        self.wdcl_connected = False
        if self.serial:
            self.serial.close()
            if self.as_interface: RNS.log((f"Closed serial port {str(self.port)} for {str(self)}"), RNS.LOG_VERBOSE) 
            else: self.owner.wlog(f"Closed serial port {str(self.port.device)} for {str(self)}")

    def configure_device(self):
        thread = threading.Thread(target=self.read_loop)
        thread.daemon = True
        thread.start()
        if self.as_interface: RNS.log(f"Serial port {self.port} is now open, discovering remote device...", RNS.LOG_VERBOSE)
        else: self.owner.wlog("Serial port "+self.port.device+" is now open")
        self.device.discover()

        if self.as_interface:
            timeout = time.time() + self.WDCL_HANDSHAKE_TIMEOUT
            while time.time() < timeout and not self.wdcl_connected: time.sleep(0.1)
            if not self.wdcl_connected:
                raise IOError(f"WDCL connection handshake timed out for {self}")
                self.online = False
                self.wdcl_connected = False
                if self.serial:
                    try: self.serial.close()
                    except Exception as e: RNS.log("Error while cleaning serial connection: {e}", RNS.LOG_ERROR)

        self.online = True

    def process_incoming(self, data):
        self.rxb += len(data)
        if self.device:
            while len(self.frame_queue): self.device.incoming_frame(self.frame_queue.pop())
            self.device.incoming_frame(data)
        else: self.frame_queue.append(data)

    def process_outgoing(self, data):
        if self.serial.is_open:
            data = bytes([HDLC.FLAG])+HDLC.escape(data)+bytes([HDLC.FLAG])
            written = self.serial.write(data)
            self.txb += len(data)          
            if written != len(data):
                raise IOError("Serial interface only wrote "+str(written)+" bytes of "+str(len(data)))

    def read_loop(self):
        try:
            while self.serial.is_open:
                data_in = self.serial.read(1500)
                if len(data_in) > 0:
                    self.frame_buffer += data_in
                    flags_remaining = True
                    while flags_remaining:
                        frame_start = self.frame_buffer.find(HDLC.FLAG)
                        if frame_start != -1:
                            frame_end = self.frame_buffer.find(HDLC.FLAG, frame_start+1)
                            if frame_end != -1:
                                frame = self.frame_buffer[frame_start+1:frame_end]
                                frame = frame.replace(bytes([HDLC.ESC, HDLC.FLAG ^ HDLC.ESC_MASK]), bytes([HDLC.FLAG]))
                                frame = frame.replace(bytes([HDLC.ESC, HDLC.ESC  ^ HDLC.ESC_MASK]), bytes([HDLC.ESC]))
                                if len(frame) > WDCL.HEADER_MINSIZE: self.process_incoming(frame)
                                self.frame_buffer = self.frame_buffer[frame_end:]
                            else:
                                flags_remaining = False
                        else:
                            flags_remaining = False
                    
        except Exception as e:
            self.online = False
            self.wdcl_connected = False
            if self.should_run:
                if self.as_interface:
                    RNS.log("A serial port error occurred, the contained exception was: "+str(e), RNS.LOG_ERROR)
                    RNS.log("Will attempt to reconnect the interface periodically.", RNS.LOG_ERROR)
                else:
                    self.owner.wlog("A serial port error occurred, the contained exception was: "+str(e))
                    self.owner.wlog("Will attempt to reconnect the interface periodically.")
                RNS.trace_exception(e)

        self.online = False
        self.wdcl_connected = False
        try: self.serial.close()
        except: pass
        if self.should_run: self.reconnect_port()

    def reconnect_port(self):
        if self.reconnecting: return
        self.reconnecting = True
        self.wdcl_connected = False
        while not self.online:
            try:
                time.sleep(5)
                if self.as_interface: RNS.log("Attempting to reconnect serial port "+str(self.port)+" for "+str(self.owner)+"...", RNS.LOG_DEBUG)
                else: self.owner.wlog("Attempting to reconnect serial port "+str(self.port.device)+" for "+str(self.owner)+"...")
                self.open_port()
                if self.serial and self.serial.is_open: self.configure_device()
            except Exception as e:
                if self.as_interface: RNS.log("Error while reconnecting port, the contained exception was: "+str(e), RNS.LOG_ERROR)
                else: self.owner.wlog("Error while reconnecting port, the contained exception was: "+str(e))
                RNS.trace_exception(e)

        self.reconnecting = False
        if self.as_interface: RNS.log("Reconnected serial port for "+str(self), RNS.LOG_INFO)
        else: self.owner.wlog("Reconnected serial port for "+str(self))

    def __str__(self):
        if self.as_interface: return f"WDCL over {self.port}"
        else:
            if self.port.serial_number: sn_str = f" {self.port.serial_number}"
            else: sn_str = ""
            return f"{self.port.product}{sn_str} (USB)"

class Cmd():
    WDCL_CMD_ENDPOINT_PKT        = 0x0001
    WDCL_CMD_ENDPOINTS_LIST      = 0x0100
    WDCL_CMD_REMOTE_DISPLAY      = 0x0A00
    WDCL_CMD_REMOTE_INPUT        = 0x0A01

class Evt():
    ET_MSG                       = 0x0000
    ET_SYSTEM_BOOT               = 0x0001
    ET_CORE_INIT                 = 0x0002
    ET_DRV_UART_INIT             = 0x1000
    ET_DRV_USB_CDC_INIT          = 0x1010
    ET_DRV_USB_CDC_HOST_AVAIL    = 0x1011
    ET_DRV_USB_CDC_HOST_SUSPEND  = 0x1012
    ET_DRV_USB_CDC_HOST_RESUME   = 0x1013
    ET_DRV_USB_CDC_CONNECTED     = 0x1014
    ET_DRV_USB_CDC_READ_ERR      = 0x1015
    ET_DRV_USB_CDC_OVERFLOW      = 0x1016
    ET_DRV_USB_CDC_DROPPED       = 0x1017
    ET_DRV_USB_CDC_TX_TIMEOUT    = 0x1018
    ET_DRV_I2C_INIT              = 0x1020
    ET_DRV_NVS_INIT              = 0x1030
    ET_DRV_NVS_ERASE             = 0x1031
    ET_DRV_CRYPTO_INIT           = 0x1040
    ET_DRV_DISPLAY_INIT          = 0x1050
    ET_DRV_DISPLAY_BUS_AVAILABLE = 0x1051
    ET_DRV_DISPLAY_IO_CONFIGURED = 0x1052
    ET_DRV_DISPLAY_PANEL_CREATED = 0x1053
    ET_DRV_DISPLAY_PANEL_RESET   = 0x1054
    ET_DRV_DISPLAY_PANEL_INIT    = 0x1055
    ET_DRV_DISPLAY_PANEL_ENABLE  = 0x1056
    ET_DRV_DISPLAY_REMOTE_ENABLE = 0x1057
    ET_DRV_W80211_INIT           = 0x1060
    ET_DRV_W80211_INIT           = 0x1061
    ET_DRV_W80211_CHANNEL        = 0x1062
    ET_DRV_W80211_POWER          = 0x1063
    ET_KRN_LOGGER_INIT           = 0x2000
    ET_KRN_LOGGER_OUTPUT         = 0x2001
    ET_KRN_UI_INIT               = 0x2010
    ET_PROTO_WDCL_INIT           = 0x3000
    ET_PROTO_WDCL_RUNNING        = 0x3001
    ET_PROTO_WDCL_CONNECTION     = 0x3002
    ET_PROTO_WDCL_HOST_ENDPOINT  = 0x3003
    ET_PROTO_WEAVE_INIT          = 0x3100
    ET_PROTO_WEAVE_RUNNING       = 0x3101
    ET_PROTO_WEAVE_EP_ALIVE      = 0x3102
    ET_PROTO_WEAVE_EP_TIMEOUT    = 0x3103
    ET_PROTO_WEAVE_EP_VIA        = 0x3104
    ET_SRVCTL_REMOTE_DISPLAY     = 0xA000
    ET_INTERFACE_REGISTERED      = 0xD000
    ET_STAT_STATE                = 0xE000
    ET_STAT_UPTIME               = 0xE001
    ET_STAT_TIMEBASE             = 0xE002
    ET_STAT_CPU                  = 0xE003
    ET_STAT_TASK_CPU             = 0xE004
    ET_STAT_MEMORY               = 0xE005
    ET_STAT_STORAGE              = 0xE006
    ET_SYSERR_MEM_EXHAUSTED      = 0xF000

    IF_TYPE_USB                  = 0x01
    IF_TYPE_UART                 = 0x02
    IF_TYPE_W80211               = 0x03
    IF_TYPE_BLE                  = 0x04
    IF_TYPE_LORA                 = 0x05
    IF_TYPE_ETHERNET             = 0x06
    IF_TYPE_WIFI                 = 0x07
    IF_TYPE_TCP                  = 0x08
    IF_TYPE_UDP                  = 0x09
    IF_TYPE_IR                   = 0x0A
    IF_TYPE_AFSK                 = 0x0B
    IF_TYPE_GPIO                 = 0x0C
    IF_TYPE_SPI                  = 0x0D
    IF_TYPE_I2C                  = 0x0E
    IF_TYPE_CAN                  = 0x0F
    IF_TYPE_DMA                  = 0x10

    event_descriptions = {
        ET_SYSTEM_BOOT: "System boot",
        ET_CORE_INIT: "Core initialization",
        ET_DRV_UART_INIT: "UART driver initialization",
        ET_DRV_USB_CDC_INIT: "USB CDC driver initialization",
        ET_DRV_USB_CDC_HOST_AVAIL: "USB CDC host became available",
        ET_DRV_USB_CDC_HOST_SUSPEND: "USB CDC host suspend",
        ET_DRV_USB_CDC_HOST_RESUME: "USB CDC host resume",
        ET_DRV_USB_CDC_CONNECTED: "USB CDC host connection",
        ET_DRV_USB_CDC_READ_ERR: "USB CDC read error",
        ET_DRV_USB_CDC_OVERFLOW: "USB CDC overflow occurred",
        ET_DRV_USB_CDC_DROPPED: "USB CDC dropped bytes",
        ET_DRV_USB_CDC_TX_TIMEOUT: "USB CDC TX flush timeout",
        ET_DRV_I2C_INIT: "I2C driver initialization",
        ET_DRV_NVS_INIT: "NVS driver initialization",
        ET_DRV_CRYPTO_INIT: "Cryptography driver initialization",
        ET_DRV_W80211_INIT: "W802.11 driver initialization",
        ET_DRV_W80211_CHANNEL: "W802.11 channel configuration",
        ET_DRV_W80211_POWER: "W802.11 TX power configuration",
        ET_DRV_DISPLAY_INIT: "Display driver initialization",
        ET_DRV_DISPLAY_BUS_AVAILABLE: "Display bus availability",
        ET_DRV_DISPLAY_IO_CONFIGURED: "Display I/O configuration",
        ET_DRV_DISPLAY_PANEL_CREATED: "Display panel allocation",
        ET_DRV_DISPLAY_PANEL_RESET: "Display panel reset",
        ET_DRV_DISPLAY_PANEL_INIT: "Display panel initialization",
        ET_DRV_DISPLAY_PANEL_ENABLE: "Display panel activation",
        ET_DRV_DISPLAY_REMOTE_ENABLE: "Remote display output activation",
        ET_KRN_LOGGER_INIT: "Logging service initialization",
        ET_KRN_LOGGER_OUTPUT: "Logging service output activation",
        ET_KRN_UI_INIT: "User interface service initialization",
        ET_PROTO_WDCL_INIT: "WDCL protocol initialization",
        ET_PROTO_WDCL_RUNNING: "WDCL protocol activation",
        ET_PROTO_WDCL_CONNECTION: "WDCL host connection",
        ET_PROTO_WDCL_HOST_ENDPOINT: "Weave host endpoint",
        ET_PROTO_WEAVE_INIT: "Weave protocol initialization",
        ET_PROTO_WEAVE_RUNNING: "Weave protocol activation",
        ET_PROTO_WEAVE_EP_ALIVE: "Weave endpoint alive",
        ET_PROTO_WEAVE_EP_TIMEOUT: "Weave endpoint disappeared",
        ET_SRVCTL_REMOTE_DISPLAY: "Remote display service control event",
        ET_INTERFACE_REGISTERED: "Interface registration",
        ET_SYSERR_MEM_EXHAUSTED: "System memory exhausted",
    }

    interface_types = {
        IF_TYPE_USB: "usb",
        IF_TYPE_UART: "uart",
        IF_TYPE_W80211: "mw",
        IF_TYPE_BLE: "ble",
        IF_TYPE_LORA: "lora",
        IF_TYPE_ETHERNET: "eth",
        IF_TYPE_WIFI: "wifi",
        IF_TYPE_TCP: "tcp",
        IF_TYPE_UDP: "udp",
        IF_TYPE_IR: "ir",
        IF_TYPE_AFSK: "afsk",
        IF_TYPE_GPIO: "gpio",
        IF_TYPE_SPI: "spi",
        IF_TYPE_I2C: "i2c",
        IF_TYPE_CAN: "can",
        IF_TYPE_DMA: "dma",
    }

    channel_descriptions = {
        1: "Channel 1 (2412 MHz)",
        2: "Channel 2 (2417 MHz)",
        3: "Channel 3 (2422 MHz)",
        4: "Channel 4 (2427 MHz)",
        5: "Channel 5 (2432 MHz)",
        6: "Channel 6 (2437 MHz)",
        7: "Channel 7 (2442 MHz)",
        8: "Channel 8 (2447 MHz)",
        9: "Channel 9 (2452 MHz)",
        10: "Channel 10 (2457 MHz)",
        11: "Channel 11 (2462 MHz)",
        12: "Channel 12 (2467 MHz)",
        13: "Channel 13 (2472 MHz)",
        14: "Channel 14 (2484 MHz)",
    }

    LOG_FORCE    = 0
    LOG_CRITICAL = 1
    LOG_ERROR    = 2
    LOG_WARNING  = 3
    LOG_NOTICE   = 4
    LOG_INFO     = 5
    LOG_VERBOSE  = 6
    LOG_DEBUG    = 7
    LOG_EXTREME  = 8
    LOG_SYSTEM   = 9

    levels = {
        LOG_FORCE: "Forced",
        LOG_CRITICAL: "Critical",
        LOG_ERROR: "Error",
        LOG_WARNING: "Warning",
        LOG_NOTICE: "Notice",
        LOG_INFO: "Info",
        LOG_VERBOSE: "Verbose",
        LOG_DEBUG: "Debug",
        LOG_EXTREME: "Extreme",
        LOG_SYSTEM: "System",
    }

    task_descriptions = {
        "taskLVGL": "Driver: UI Renderer",
        "ui_service": "Service: User Interface",
        "TinyUSB":  "Driver: USB",
        "drv_w80211": "Driver: W802.11",
        "system_stats": "System: Stats",
        "core": "System: Core",
        "protocol_wdcl": "Protocol: WDCL",
        "protocol_weave": "Protocol: Weave",
        "tiT": "Protocol: TCP/IP",
        "ipc0": "System: CPU 0 IPC",
        "ipc1": "System: CPU 1 IPC",
        "esp_timer": "Driver: Timers",
        "Tmr Svc": "Service: Timers",
        "kernel_logger": "Service: Logging",
        "remote_display": "Service: Remote Display",
        "wifi": "System: WiFi Hardware",
        "sys_evt": "System: Kernel Events",
    }

    @staticmethod
    def level(level):
        if level in Evt.levels: return Evt.levels[level]
        else: return "Unknown"

class LogFrame():
    timestamp = None
    level = None
    event = None
    data = b""

    def __init__(self, timestamp=None, level=None, event=None, data=b""):
        self.timestamp = timestamp; self.level = level
        self.event = event; self.data = data

class WeaveEndpoint():
    QUEUE_LEN = 1024

    def __init__(self, endpoint_addr):
        self.endpoint_addr = endpoint_addr
        self.alive = time.time()
        self.via = None
        self.received = deque(maxlen=WeaveEndpoint.QUEUE_LEN)

    def receive(self, data):
        self.received.append(data)

class WeaveDevice():
    STATLEN_MAX            = 120
    STAT_UPDATE_THROTTLE   = 0.5

    WEAVE_SWITCH_ID_LEN    = 4
    WEAVE_ENDPOINT_ID_LEN  = 8
    WEAVE_FLOWSEQ_LEN      = 2
    WEAVE_HMAC_LEN         = 8
    WEAVE_AUTH_LEN         = WEAVE_ENDPOINT_ID_LEN+WEAVE_HMAC_LEN

    WEAVE_PUBKEY_SIZE      = 32
    WEAVE_PRVKEY_SIZE      = 64
    WEAVE_SIGNATURE_LEN    = 64

    def __init__(self, as_interface=False, rns_interface=None):
        self.identity        = None
        self.receiver        = None
        self.switch_id       = None
        self.endpoint_id     = None
        self.owner           = None
        self.rns_interface   = rns_interface
        self.as_interface    = as_interface
        self.endpoints       = {}
        self.active_tasks    = {}
        self.cpu_load        = 0
        self.memory_total    = 0
        self.memory_free     = 0
        self.memory_used     = 0
        self.memory_used_pct = 0
        self.log_queue       = deque()
        self.memory_stats    = deque(maxlen=WeaveDevice.STATLEN_MAX)
        self.cpu_stats       = deque(maxlen=WeaveDevice.STATLEN_MAX)
        self.display_buffer  = bytearray(0)
        self.update_display  = False

        self.next_update_memory = 0
        self.next_update_cpu    = 0

    def wdcl_send(self, packet_type, data):
        if not self.switch_id:
            if self.as_interface: RNS.log("Attempt to transmit on {self} while remote Weave device identity is unknown", RNS.LOG_ERROR)
            else: self.receiver.log("Error: Attempt to transmit while remote Weave device identity is unknown")
        else:
            frame  = self.switch_id
            frame += bytes([packet_type])
            frame += data
            self.connection.process_outgoing(frame)

    def wdcl_broadcast(self, packet_type, data):
        frame  = WDCL.WDCL_BROADCAST
        frame += bytes([packet_type])
        frame += data
        self.connection.process_outgoing(frame)

    def wdcl_send_command(self, command, data):
        frame  = b""
        frame += bytes([command>>8, (command & 0xFF)])
        frame += data
        self.wdcl_send(WDCL.WDCL_T_CMD, frame)

    def discover(self):
        self.wdcl_broadcast(WDCL.WDCL_T_DISCOVER, self.connection.switch_id)

    def handshake(self):
        if self.identity == None:
            if self.as_interface: RNS.log("Attempt to perform handshake on {self} before remote device discovery completion", RNS.LOG_ERROR)
            else: self.receiver.log("Attempt to perform handshake before remote device discovery completion")
        else:
            signed_id = self.switch_id
            signature = self.connection.switch_identity.sign(signed_id)
            data      = self.connection.switch_pub_bytes
            data     += signature
            self.wdcl_send(WDCL.WDCL_T_CONNECT, data)
            if self.as_interface: RNS.log(f"WDCL connection handshake sent", RNS.LOG_VERBOSE)
            else: self.receiver.log("Connection handshake sent")

    def capture_stats_cpu(self):
        self.cpu_stats.append({"timestamp": time.time(), "cpu_load": self.cpu_load})
        if self.receiver and self.receiver.ready and len(self.memory_stats) > 1: self.receiver.stats_update("cpu")

    def capture_stats_memory(self):
        self.memory_stats.append({"timestamp": time.time(), "memory_used": self.memory_used})
        if self.receiver and self.receiver.ready and len(self.memory_stats) > 1: self.receiver.stats_update("memory")

    def get_cpu_stats(self):
        tbegin = None
        stats = {"timestamps": [], "values": [], "max": 100, "unit": "%"}
        for i in range(0, len(self.cpu_stats)):
            if tbegin == None: tbegin = self.cpu_stats[len(self.cpu_stats)-1]["timestamp"]
            stats["timestamps"].append(self.cpu_stats[i]["timestamp"]-tbegin)
            stats["values"].append(self.cpu_stats[i]["cpu_load"])

        return stats

    def get_memory_stats(self):
        tbegin = None
        stats = {"timestamps": [], "values": [], "max": self.memory_total, "unit": "B"}
        for i in range(0, len(self.memory_stats)):
            if tbegin == None: tbegin = self.memory_stats[len(self.memory_stats)-1]["timestamp"]
            stats["timestamps"].append(self.memory_stats[i]["timestamp"]-tbegin)
            stats["values"].append(self.memory_stats[i]["memory_used"])

        return stats

    def get_active_tasks(self):
        active_tasks = {}
        now = time.time()
        for task_id in self.active_tasks:
            if not task_id.startswith("IDLE"):
                task_description = task_id
                if task_id in Evt.task_descriptions: task_description = Evt.task_descriptions[task_id]
                if now - self.active_tasks[task_id]["timestamp"] < 5:
                    active_tasks[task_description] = self.active_tasks[task_id]

        return active_tasks

    def disconnect_display(self):
        self.wdcl_send_command(Cmd.WDCL_CMD_REMOTE_DISPLAY, bytes([0x00]))
        self.update_display = False

    def connect_display(self):
        self.wdcl_send_command(Cmd.WDCL_CMD_REMOTE_DISPLAY, bytes([0x01]))
        self.update_display = True

    def endpoint_alive(self, endpoint_id):
        if not endpoint_id in self.endpoints: self.endpoints[endpoint_id] = WeaveEndpoint(endpoint_id)
        else: self.endpoints[endpoint_id].alive = time.time()

        if self.as_interface: self.rns_interface.add_peer(endpoint_id)

    def endpoint_via(self, endpoint_id, via_switch_id):
        if endpoint_id in self.endpoints: self.endpoints[endpoint_id].via = via_switch_id
        if self.as_interface: self.rns_interface.endpoint_via(endpoint_id, via_switch_id)

    def deliver_packet(self, endpoint_id, data):
        packet_data = endpoint_id+data
        self.wdcl_send_command(Cmd.WDCL_CMD_ENDPOINT_PKT, packet_data)

    def received_packet(self, source, data):
        self.endpoint_alive(source)
        if self.as_interface:
            self.rns_interface.process_incoming(data, source)

    def incoming_frame(self, data):
        if len(data) > self.WEAVE_SWITCH_ID_LEN+2 and data[self.WEAVE_SWITCH_ID_LEN] == WDCL.WDCL_T_ENDPOINT_PKT and data[:self.WEAVE_SWITCH_ID_LEN] == self.connection.switch_id:
            payload = data[self.WEAVE_SWITCH_ID_LEN+1:-self.WEAVE_ENDPOINT_ID_LEN]
            src_endpoint = data[-self.WEAVE_ENDPOINT_ID_LEN:]
            self.received_packet(src_endpoint, payload)

        elif len(data) > self.WEAVE_SWITCH_ID_LEN+1 and data[self.WEAVE_SWITCH_ID_LEN] == WDCL.WDCL_T_DISCOVER:
            discovery_response_len = self.WEAVE_SWITCH_ID_LEN+1+self.WEAVE_PUBKEY_SIZE+self.WEAVE_SIGNATURE_LEN
            if len(data) == discovery_response_len:
                signed_id        = data[:self.WEAVE_SWITCH_ID_LEN]
                remote_pub_key   = data[self.WEAVE_SWITCH_ID_LEN+1:self.WEAVE_SWITCH_ID_LEN+1+self.WEAVE_PUBKEY_SIZE]
                remote_switch_id = remote_pub_key[-4:]
                remote_signature = data[self.WEAVE_SWITCH_ID_LEN+1+self.WEAVE_PUBKEY_SIZE:self.WEAVE_SWITCH_ID_LEN+1+self.WEAVE_PUBKEY_SIZE+self.WEAVE_SIGNATURE_LEN]
                remote_identity = RNS.Identity(create_keys=False)
                remote_identity.load_public_key(remote_pub_key*2)
                if remote_identity.validate(remote_signature, signed_id):
                    if self.as_interface: RNS.log(f"Remote Weave device {RNS.hexrep(remote_switch_id)} discovered", RNS.LOG_VERBOSE)
                    else: self.receiver.log(f"Remote Weave device {RNS.hexrep(remote_switch_id)} discovered")
                    self.identity = remote_identity
                    self.switch_id = remote_switch_id
                    self.handshake()
                else:
                    if self.as_interface: RNS.LOG("Invalid remote device discovery response received", RNS.LOG_ERROR)
                    else: self.receiver.log("Invalid remote device discovery response received")

        elif len(data) > self.WEAVE_SWITCH_ID_LEN+1 and data[self.WEAVE_SWITCH_ID_LEN] == WDCL.WDCL_T_LOG:
            fd  = data[self.WEAVE_SWITCH_ID_LEN+2:]
            ts  = fd[1] << 24 | fd[2] << 16 | fd[3] << 8 | fd[4]
            lvl = fd[5]; evt = fd[6] << 8 | fd[7]; data = fd[8:]
            self.log_handle(LogFrame(timestamp=ts/1000.0, level=lvl, event=evt, data=data))

        elif len(data) > self.WEAVE_SWITCH_ID_LEN+10 and data[self.WEAVE_SWITCH_ID_LEN] == WDCL.WDCL_T_DISP:
            fd  = data[self.WEAVE_SWITCH_ID_LEN+1:]
            cf  = fd[0]
            ofs = fd[1] << 24 | fd[2] << 16 | fd[3] << 8 | fd[4]
            dsz = fd[5] << 24 | fd[6] << 16 | fd[7] << 8 | fd[8]
            fbf = fd[9:]

            w = 128; h = 64

            if dsz > len(self.display_buffer): self.display_buffer = bytearray(dsz)
            self.display_buffer[ofs:ofs+len(fbf)] = fbf

            if self.receiver and self.receiver.ready and ofs+len(fbf) == dsz:
                if self.update_display: self.receiver.display_update(self.display_buffer, w, h)

    def log_handle(self, frame):
        # Handle system event signalling
        if frame.event == Evt.ET_PROTO_WDCL_CONNECTION: self.connection.wdcl_connected = True
        if frame.event == Evt.ET_PROTO_WDCL_HOST_ENDPOINT and len(frame.data) == self.WEAVE_ENDPOINT_ID_LEN: self.endpoint_id = frame.data
        if frame.event == Evt.ET_PROTO_WEAVE_EP_ALIVE and len(frame.data) == self.WEAVE_ENDPOINT_ID_LEN: self.endpoint_alive(frame.data)
        if frame.event == Evt.ET_PROTO_WEAVE_EP_VIA and len(frame.data) == self.WEAVE_ENDPOINT_ID_LEN+self.WEAVE_SWITCH_ID_LEN: self.endpoint_via(frame.data[:self.WEAVE_ENDPOINT_ID_LEN], frame.data[self.WEAVE_ENDPOINT_ID_LEN:])
        elif frame.event == Evt.ET_STAT_TASK_CPU: self.active_tasks[frame.data[1:].decode("utf-8")] = { "cpu_load": frame.data[0], "timestamp": time.time() }
        elif frame.event == Evt.ET_STAT_CPU: 
            self.cpu_load = frame.data[0]
            self.capture_stats_cpu()
        elif frame.event == Evt.ET_STAT_MEMORY:
            self.memory_free     = int.from_bytes(frame.data[:4])
            self.memory_total    = int.from_bytes(frame.data[4:])
            self.memory_used     = self.memory_total-self.memory_free
            self.memory_used_pct = round((self.memory_used/self.memory_total)*100, 2)
            self.capture_stats_memory()

        # Handle generic messages and unmapped events
        else:
            ts = RNS.prettytime(frame.timestamp)
            if frame.event == Evt.ET_MSG:
                if len(frame.data): data_string = frame.data.decode("utf-8")
                else: data_string = ""
                rendered = f"[{ts}] [{Evt.level(frame.level)}]: {data_string}"

            else:
                if frame.event in Evt.event_descriptions: event_description = Evt.event_descriptions[frame.event]
                else: event_description = f"0x{RNS.hexrep(frame.event, delimit=False)}"

                if frame.event == Evt.ET_INTERFACE_REGISTERED:
                    if len(frame.data) >= 2:
                        interface_index = frame.data[0]; interface_type = frame.data[1]
                        type_name = "phy"
                        if interface_type in Evt.interface_types: type_name = Evt.interface_types[interface_type]
                        data_string = f": {type_name}{interface_index}"
                    else: data_string = ""
                else:
                    if len(frame.data):
                        data_string = f": {RNS.hexrep(frame.data)}"
                        if   frame.event == Evt.ET_DRV_USB_CDC_CONNECTED:
                            if   frame.data[0] == 0x01: data_string = ": Connected"
                            elif frame.data[0] == 0x00: data_string = ": Disconnected"
                        elif frame.event == Evt.ET_DRV_W80211_CHANNEL:
                            if   frame.data[0] in Evt.channel_descriptions: data_string = f": {Evt.channel_descriptions[frame.data[0]]}"
                            else:                                           data_string = f": {RNS.hexrep(frame.data)}"
                        elif frame.event == Evt.ET_DRV_W80211_POWER:
                            tx_power = frame.data[0]*0.25
                            data_string = f": {tx_power} dBm ({int(10**(tx_power/10))} mW)"
                        elif frame.event >= Evt.ET_CORE_INIT and frame.event <= Evt.ET_PROTO_WEAVE_RUNNING:
                            if   frame.data[0] == 0x01: data_string = ": Success"
                            elif frame.data[0] == 0x00:
                                if frame.level == Evt.LOG_ERROR: data_string = ": Failure"
                                else:                            data_string = ": Stopped"
                            else:                       data_string = f": {RNS.hexrep(frame.data)}"
                        
                    else: data_string = ""

                rendered = f"[{ts}] [{Evt.level(frame.level)}] [{event_description}]{data_string}"
            
            if self.as_interface:
                RNS.log(f"{self.rns_interface}: {rendered}", RNS.LOG_EXTREME)
            else:
                if self.receiver and self.receiver.ready:
                    while len(self.log_queue): self.receiver.log(self.log_queue.pop())
                    self.receiver.log(rendered)
                else: self.log_queue.append(rendered)

class WeaveInterface(Interface):
    HW_MTU = 1024
    FIXED_MTU = True

    DEFAULT_IFAC_SIZE  = 16
    PEERING_TIMEOUT    = 20.0
    BITRATE_GUESS      = 250*1000

    MULTI_IF_DEQUE_LEN = 48
    MULTI_IF_DEQUE_TTL = 0.75

    @property
    def cpu_load(self):
        if not self.device: return None
        else: return self.device.cpu_load

    @property
    def mem_load(self):
        if not self.device: return None
        else: return self.device.memory_used_pct

    @property
    def switch_id(self):
        if not self.device: return None
        else: return self.device.switch_id

    @property
    def endpoint_id(self):
        if not self.device: return None
        else: return self.device.endpoint_id

    def __init__(self, owner, configuration):
        c                      = Interface.get_config_obj(configuration)
        name                   = c["name"]
        port                   = c["port"]
        configured_bitrate     = c["configured_bitrate"] if "configured_bitrate" in c else None

        from RNS.Interfaces import netinfo
        super().__init__()
        self.netinfo = netinfo

        self.HW_MTU = WeaveInterface.HW_MTU
        self.IN  = True
        self.OUT = False
        self.name = name
        self.port = port
        self.switch_identity = RNS.Identity()
        self.owner = owner
        self.hw_errors = []
        self._online = False
        self.final_init_done = False
        self.peers = {}
        self.timed_out_interfaces = {}
        self.spawned_interfaces = {}
        self.write_lock = threading.Lock()
        self.mif_deque = deque(maxlen=WeaveInterface.MULTI_IF_DEQUE_LEN)
        self.mif_deque_times = deque(maxlen=WeaveInterface.MULTI_IF_DEQUE_LEN)

        self.announce_rate_target = None
        self.peer_job_interval = WeaveInterface.PEERING_TIMEOUT*1.1
        self.peering_timeout   = WeaveInterface.PEERING_TIMEOUT

        self.receives = True
        if configured_bitrate != None: self.bitrate = configured_bitrate
        else: self.bitrate = WeaveInterface.BITRATE_GUESS

    def final_init(self):
        self.device = WeaveDevice(as_interface=True, rns_interface=self)
        self.connection = WDCL(owner=self, device=self.device, port=self.port, as_interface=True)

        job_thread = threading.Thread(target=self.peer_jobs)
        job_thread.daemon = True
        job_thread.start()

        self._online = True
        self.final_init_done = True

    def peer_jobs(self):
        while True:
            time.sleep(self.peer_job_interval)
            now = time.time()
            timed_out_peers = []

            # Check for timed out peers
            for peer_addr in self.peers:
                peer = self.peers[peer_addr]
                last_heard = peer[1]
                if now > last_heard+self.peering_timeout:
                    timed_out_peers.append(peer_addr)

            # Remove any timed out peers
            for peer_addr in timed_out_peers:
                removed_peer = self.peers.pop(peer_addr)
                if peer_addr in self.spawned_interfaces:
                    spawned_interface = self.spawned_interfaces[peer_addr]
                    spawned_interface.detach()
                    spawned_interface.teardown()
                RNS.log(str(self)+" removed peer "+RNS.hexrep(peer_addr)+" on "+RNS.hexrep(removed_peer[0]), RNS.LOG_DEBUG)
            
    @property
    def peer_count(self):
        return len(self.spawned_interfaces)

    def endpoint_via(self, endpoint_addr, via_switch_addr):
        if endpoint_addr in self.peers: self.peers[endpoint_addr][2].via_switch_id = via_switch_addr

    def add_peer(self, endpoint_addr):    
        if not endpoint_addr in self.peers:
            spawned_interface = WeaveInterfacePeer(self, endpoint_addr)
            spawned_interface.OUT = self.OUT
            spawned_interface.IN  = self.IN

            spawned_interface.ingress_control = self.ingress_control
            spawned_interface.ic_max_held_announces = self.ic_max_held_announces
            spawned_interface.ic_burst_hold = self.ic_burst_hold
            spawned_interface.ic_burst_freq = self.ic_burst_freq
            spawned_interface.ic_burst_freq_new = self.ic_burst_freq_new
            spawned_interface.ic_new_time = self.ic_new_time
            spawned_interface.ic_burst_penalty = self.ic_burst_penalty
            spawned_interface.ic_held_release_interval = self.ic_held_release_interval
            
            spawned_interface.parent_interface = self
            spawned_interface.bitrate = self.bitrate
            
            spawned_interface.ifac_size = self.ifac_size
            spawned_interface.ifac_netname = self.ifac_netname
            spawned_interface.ifac_netkey = self.ifac_netkey
            if spawned_interface.ifac_netname != None or spawned_interface.ifac_netkey != None:
                ifac_origin = b""
                if spawned_interface.ifac_netname != None:
                    ifac_origin += RNS.Identity.full_hash(spawned_interface.ifac_netname.encode("utf-8"))
                if spawned_interface.ifac_netkey != None:
                    ifac_origin += RNS.Identity.full_hash(spawned_interface.ifac_netkey.encode("utf-8"))

                ifac_origin_hash = RNS.Identity.full_hash(ifac_origin)
                spawned_interface.ifac_key = RNS.Cryptography.hkdf(
                    length=64,
                    derive_from=ifac_origin_hash,
                    salt=RNS.Reticulum.IFAC_SALT,
                    context=None
                )
                spawned_interface.ifac_identity = RNS.Identity.from_bytes(spawned_interface.ifac_key)
                spawned_interface.ifac_signature = spawned_interface.ifac_identity.sign(RNS.Identity.full_hash(spawned_interface.ifac_key))

            spawned_interface.announce_rate_target = self.announce_rate_target
            spawned_interface.announce_rate_grace = self.announce_rate_grace
            spawned_interface.announce_rate_penalty = self.announce_rate_penalty
            spawned_interface.mode = self.mode
            spawned_interface.HW_MTU = self.HW_MTU
            spawned_interface._online = True
            RNS.Transport.add_interface(spawned_interface)
            if endpoint_addr in self.spawned_interfaces:
                self.spawned_interfaces[endpoint_addr].detach()
                self.spawned_interfaces[endpoint_addr].teardown()
                self.spawned_interfaces.pop(spawned_interface)

            self.spawned_interfaces[endpoint_addr] = spawned_interface
            self.peers[endpoint_addr] = [endpoint_addr, time.time(), spawned_interface]

            RNS.log(f"{self} added peer {RNS.hexrep(endpoint_addr)}", RNS.LOG_DEBUG)
        else:
            self.refresh_peer(endpoint_addr)

    def refresh_peer(self, endpoint_addr):
        try:
            self.peers[endpoint_addr][1] = time.time()
        except Exception as e:
            RNS.log(f"An error occurred while refreshing peer {RNS.hexrep(endpoint_addr)} on {self}: {e}", RNS.LOG_ERROR)

    def process_incoming(self, data, endpoint_addr=None):
        if self.online and endpoint_addr in self.spawned_interfaces:
            self.spawned_interfaces[endpoint_addr].process_incoming(data, endpoint_addr)

    def process_outgoing(self,data):
        pass

    def detach(self):
        self._online = False

    @property
    def online(self):
        if not self._online: return False
        else: return self.connection.online

    @online.setter
    def online(self, value):
        self._online = value
    
    def __str__(self):
        return "WeaveInterface["+self.name+"]"

class WeaveInterfacePeer(Interface):

    def __init__(self, owner, endpoint_addr):
        super().__init__()
        self.owner = owner
        self.parent_interface = owner
        self.endpoint_addr = endpoint_addr
        self.via_switch_id = None
        self.peer_addr = None
        self.addr_info = None
        self.HW_MTU = self.owner.HW_MTU
        self.FIXED_MTU = self.owner.FIXED_MTU
        self._online = False

    def __str__(self):
        return f"WeaveInterfacePeer[{RNS.hexrep(self.endpoint_addr)}]"

    @property
    def online(self):
        if not self._online or not self.owner: return false
        else: return self.owner.online

    @online.setter
    def online(self, value):
        self._online = value

    def process_incoming(self, data, endpoint_addr=None):
        if self.online:
            data_hash = RNS.Identity.full_hash(data)
            deque_hit = False
            if data_hash in self.owner.mif_deque:
                for te in self.owner.mif_deque_times:
                    if te[0] == data_hash and time.time() < te[1]+WeaveInterface.MULTI_IF_DEQUE_TTL:
                        deque_hit = True
                        break

            if not deque_hit:
                self.owner.refresh_peer(self.endpoint_addr)
                self.owner.mif_deque.append(data_hash)
                self.owner.mif_deque_times.append([data_hash, time.time()])
                self.rxb += len(data)
                self.owner.rxb += len(data)
                self.owner.owner.inbound(data, self)

    def process_outgoing(self, data):
        if self.online:
            with self.owner.write_lock:
                try:
                    self.owner.device.deliver_packet(self.endpoint_addr, data)
                    self.txb += len(data)
                    self.owner.txb += len(data)
                except Exception as e:
                    RNS.log("Could not transmit on "+str(self)+". The contained exception was: "+str(e), RNS.LOG_ERROR)

    def detach(self):
        self._online = False
        self.detached = True
        
    def teardown(self):
        if not self.detached:
            RNS.log("The interface "+str(self)+" experienced an unrecoverable error and is being torn down.", RNS.LOG_ERROR)
            if RNS.Reticulum.panic_on_interface_error:
                RNS.panic()

        else: RNS.log("The interface "+str(self)+" is being torn down.", RNS.LOG_VERBOSE)

        self._online = False
        self.OUT = False
        self.IN = False

        if self.endpoint_addr in self.owner.spawned_interfaces:
            try: self.owner.spawned_interfaces.pop(self.endpoint_addr)
            except Exception as e:
                RNS.log(f"Could not remove {self} from parent interface on detach. The contained exception was: {e}", RNS.LOG_ERROR)

        RNS.Transport.remove_interface(self)
